from .url import *

from .basic_authorization import *
from .compressors import *
from .constants import *
from .content_type import *
from .cookiejar import *
from .exceptions import *
from .form_data import *
from .header_building_and_parsing import *
from .headers import *
from .helpers import *
from .http_message import *
from .http_protocol import *
from .http_stream_writer import *
from .keep_alive_info import *
from .multipart import *
from .quoting import *
from .web_socket_frame import *

from . import headers


__all__ = (
    *url.__all__,
    
    *basic_authorization.__all__,
    *compressors.__all__,
    *constants.__all__,
    *content_type.__all__,
    *cookiejar.__all__,
    *exceptions.__all__,
    *form_data.__all__,
    *header_building_and_parsing.__all__,
    *headers.__all__,
    *helpers.__all__,
    *http_message.__all__,
    *http_protocol.__all__,
    *http_stream_writer.__all__,
    *keep_alive_info.__all__,
    *multipart.__all__,
    *quoting.__all__,
    *web_socket_frame.__all__,
    
    'headers',
)


# Deprecations

from warnings import warn

def __getattr__(attribute_name):
    if attribute_name == 'Formdata':
        warn(
            (
                f'`Formdata` has been renamed to `FormData`.'
                f'`Formdata` will be removed in 2025 August.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        return FormData
    
    if attribute_name == 'BasicAuth':
        warn(
            (
                f'`BasicAuth` has been renamed to `BasicAuthorization`.'
                f'`BasicAuth` will be removed in 2025 November.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        return BasicAuthorization
    
    raise AttributeError(attribute_name)



__all__ = ()

from ..utils import un_map_pack


KEY_KEYBOARD_INTERRUPT = '\x03'
KEY_DESTROY = '\x04'
KEY_TAB = '\x09'
KEY_NEW_LINE_ALL = ('\x0a', '\x0d')
KEY_ARROW_N_0 = '\x1b'
KEY_ARROW_N_1 = '\x5b'
KEY_ARROW_UP = KEY_ARROW_N_0 + KEY_ARROW_N_1 + '\x41'
KEY_ARROW_DOWN = KEY_ARROW_N_0 + KEY_ARROW_N_1 + '\x42'
KEY_ARROW_RIGHT = KEY_ARROW_N_0 + KEY_ARROW_N_1 + '\x43'
KEY_ARROW_LEFT = KEY_ARROW_N_0 + KEY_ARROW_N_1 + '\x44'
KEY_ARROW_END = KEY_ARROW_N_0 + KEY_ARROW_N_1 + '\x46'
KEY_ARROW_HOME = KEY_ARROW_N_0 + KEY_ARROW_N_1 + '\x48'
KEY_BACK_TAB = KEY_ARROW_N_0 + KEY_ARROW_N_1 + '\x5a'
KEY_DELETE_RIGHT = KEY_ARROW_N_0 + KEY_ARROW_N_1 + '\x33' + '\x7e'
KEY_DELETE_LEFT = '\x7f'


KEY_NAMES = {
    KEY_KEYBOARD_INTERRUPT: 'keyboard interrupt',
    KEY_DESTROY: 'destroy',
    
    KEY_TAB: 'tab',
    KEY_BACK_TAB: 'back tab',
    
    **un_map_pack(((character, 'new line') for character in KEY_NEW_LINE_ALL)),
    
    KEY_ARROW_UP: 'arrow up',
    KEY_ARROW_DOWN: 'arrow down',
    KEY_ARROW_RIGHT: 'arrow right',
    KEY_ARROW_LEFT: 'arrow left',
    
    KEY_ARROW_END: 'end',
    KEY_ARROW_HOME: 'home',
    
    KEY_DELETE_RIGHT: 'delete',
    KEY_DELETE_LEFT: 'backspace',
}



__all__ = ('ExceptionProxyRich',)

from ..exception_representation import get_exception_representation
from ..frame_grouping import group_frames
from ..frame_proxy import get_exception_frames

from .exception_proxy_base import ExceptionProxyBase


class ExceptionProxyRich(ExceptionProxyBase):
    """
    Rich exception proxy type.
    
    Attributes
    ----------
    exception_representation : `None | ExceptionRepresentationBase`
        The exception to proxy.
    frame_groups : `None | list<FrameGroup>`
        Cache field for the exception's frames in a grouped form.
    """
    __slots__ = ('exception_representation', 'frame_groups')
    
    def __new__(cls, exception):
        """
        Creates a new exception proxy.
        
        Parameters
        ----------
        exception : ``BaseException``
            The exception to proxy.
        """
        frames = get_exception_frames(exception)
        frame_groups = group_frames(frames)
        
        exception_representation = get_exception_representation(exception, frames)
        
        self = object.__new__(cls)
        self.exception_representation = exception_representation
        self.frame_groups = frame_groups
        return self



__all__ = ('exists_matching_variable_name',)

from .helpers_variables import _get_variable_names


def exists_matching_variable_name(frame, name):
    """
    Returns whether a variable with the same name exists.
    
    Parameters
    ----------
    frame : `None | FrameProxyBase`
        Frame to get variable names from.
    name : `str`
        Name to check for.
    
    Returns
    -------
    exists : `bool`
    """
    variable_names = _get_variable_names(frame)
    return name in variable_names



__all__ = ()

from ....utils import create_ansi_format_code


COMMAND_DOWN = '\n'
COMMAND_UP = '\u001b[F'
# COMMAND_START_LINE = '\u001b[1000D'
COMMAND_START_LINE = '\r'
COMMAND_FORMAT_RESET = create_ansi_format_code()


COMMAND_CLEAR_SCREEN_FROM_CURSOR = '\u001b[0J'
COMMAND_CLEAR_SCREEN_TILL_CURSOR = '\u001b[1J'
COMMAND_CLEAR_SCREEN_WHOLE = '\u001b[2J'

COMMAND_CLEAR_LINE_FROM_CURSOR = '\u001b[0K'
COMMAND_CLEAR_LINE_TILL_CURSOR = '\u001b[1K'
COMMAND_CLEAR_LINE_WHOLE = '\u001b[2K'


def is_command(part):
    """
    Returns whether the given part is an ansi command.
    Designed for general color commands and not for various.
    
    Parameters
    ----------
    part : `str`
        the part to check.
    
    Returns
    -------
    is_ansi_command : `bool`
    """
    return part.startswith('\u001b[')


def create_command_up(amount):
    """
    Creates a command that moves the cursor up.
    
    Parameters
    ----------
    amount : `str`
        The amount to move the cursor by.
    
    Returns
    -------
    command : `str`
    """
    if amount <= 0:
        return ''
    
    return f'\u001b[{amount}A'


def create_command_down(amount):
    """
    Creates a command that moves the cursor down.
    
    Parameters
    ----------
    position : `str`
        The amount to move the cursor by.
    
    Parameters
    ----------
    amount : `str`
        The amount to move the cursor by.
    """
    if amount <= 0:
        return ''
    
    if amount == 1:
        return COMMAND_DOWN
    
    return f'\u001b[{amount}B'


def create_command_right(amount):
    """
    Creates a command that moves the cursor right.
    
    Parameters
    ----------
    amount : `str`
        The amount to move the cursor by.
    
    Returns
    -------
    command : `str`
    """
    if amount <= 0:
        return ''
    
    return f'\u001b[{amount}C'


def create_command_left(amount):
    """
    Creates a command that moves the cursor left.
    
    Parameters
    ----------
    amount : `str`
        The amount to move the cursor by.
    
    Returns
    -------
    command : `str`
    """
    if amount <= 0:
        return ''
    
    return f'\u001b[{amount}D'



__all__ = ('HTTPClient', )

from ssl import SSLContext, create_default_context as create_default_ssl_context
from warnings import warn

from ..utils import IgnoreCaseMultiValueDictionary, RichAttributeErrorBaseType, export
from ..web_common import CookieJar, URL
from ..web_common.headers import (
    AUTHORIZATION, CONTENT_LENGTH, LOCATION, METHOD_DELETE, METHOD_GET, METHOD_HEAD, METHOD_OPTIONS, METHOD_PATCH,
    METHOD_POST, METHOD_PUT, URI
)
from ..web_common.helpers import Timeout, set_tcp_nodelay
from ..web_socket import WebSocketClient

from .client_request import ClientRequest
from .connector_tcp import ConnectorTCP
from .constants import REQUEST_TIMEOUT_DEFAULT
from .proxy import Proxy
from .request_context_manager import RequestContextManager
from .ssl_fingerprint import SSLFingerprint
from .web_socket_context_manager import WebSocketContextManager


@export
class HTTPClient(RichAttributeErrorBaseType):
    """
    HTTP client implementation.
    
    Attributes
    ----------
    connector : ``ConnectorBase``
        Connector of the http client. Defaults to ``ConnectorTCP``.
    
    cookie_jar : ``CookieJar``
        Cookies stored by the http client.
    
    loop : ``EventThread``
        The event loop used by the http client.
    
    proxy : `None | Proxy`
        Proxy to use for each request.
    """
    __slots__ = ('connector', 'cookie_jar', 'loop', 'proxy')
    
    def __new__(
        cls,
        loop,
        *deprecated,
        connector = None,
        proxy = ...,
        proxy_headers = ...,
        proxy_url = ...,
        proxy_auth = ...,
    ):
        """
        Creates a new http client with the given parameters.
        
        Parameters
        ----------
        loop : ``EventThread``
            The event loop used by the http client.
        
        connector : `None | ConnectorBase` = `None`, Optional (Keyword only)
            Connector to be used by the client.
            If not given or given as `None`, a new ``ConnectorTCP`` is created and used.
        
        proxy : `None | Proxy`, Optional
            Proxy to use with every request.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        """
        # deprecated
        deprecated_length = len(deprecated)
        if deprecated_length:
            proxy_url = deprecated[0]
            
            if deprecated_length > 1:
                proxy_auth = deprecated[1]
        
        if (proxy_auth is not ...):
            warn(
                (
                    f'`{cls.__name__}.__new__`\'s `proxy_auth` is deprecated '
                    f'and will be removed at 2025 November. '
                    'Please use `proxy = Proxy(url, authorization = authorization)` instead.'
                ),
                FutureWarning,
                stacklevel = 2,
            )
        
        if (proxy_headers is not ...):
            warn(
                (
                    f'`{cls.__name__}.__new__`\'s `proxy_headers` is deprecated '
                    f'and will be removed at 2025 November. '
                    'Please use `proxy = Proxy(url, headers = headers)` instead.'
                ),
                FutureWarning,
                stacklevel = 2,
            )
        
        if (proxy_url is not ...):
            warn(
                (
                    f'`{cls.__name__}.__new__`\'s `proxy_url` is deprecated '
                    f'and will be removed at 2025 November. '
                    'Please use `proxy = Proxy(url)` instead.'
                ),
                FutureWarning,
                stacklevel = 2,
            )
        
        if proxy is ...:
            if (proxy_url is ...):
                proxy = None
            else:
                proxy = Proxy(proxy_url, headers = proxy_headers, authorization = proxy_auth)
        
        elif (proxy is not None) and (not isinstance(proxy, Proxy)):
            raise TypeError(
                f'`proxy` can be `None`, `{Proxy.__name__}`, got '
                f'{type(proxy).__name__}; {proxy!r}.'
            )
        
        
        # connector
        if (connector is None):
            connector = ConnectorTCP(loop)
        
        
        # Construct
        self = object.__new__(cls)
        self.connector = connector
        self.cookie_jar = CookieJar()
        self.loop = loop
        self.proxy = proxy
        return self
    
    
    async def _request(self, method, url, headers, *, data = None, params = ..., query = None, redirects = 3):
        """
        Internal method for executing an http request.
        
        This method is a coroutine.
        
        Parameters
        ----------
        method : `str`
            The method of the request.
        
        url : `str | URL`
            The url to request.
        
        headers : `None | dict<str, str> | IgnoreCaseMultiValueDictionary`
            Request headers.
        
        data : `None | object` = `None`, Optional (Keyword only)
            Data to send a the body of the request.
        
        query : `None | str | dict<None, str | bool | int | float, iterable<...>> | iterable<...>` = `None`, \
                Optional (Keyword only)
            Query string parameters
        
        redirects : `int` = `3`, Optional (Keyword only)
            The maximal amount of allowed redirects.
        
        Returns
        -------
        response : ``ClientResponse``
        
        Raises
        ------
        ConnectionError
            - Too many redirects.
            - Connector closed.
        ValueError
            - Host could not be detected from `url`.
            - `compression` and `Content-Encoding` would be set at the same time.
            - `chunked` cannot be set, because `Transfer-Encoding: chunked` is already set.
            - `chunked` cannot be set, because `Content-Length` header is already present.
        OSError
            If a system function returns a system-related error.
        RuntimeError
            - If one of `data`'s field's content has unknown content-encoding.
            - If one of `data`'s field's content has unknown content-transfer-encoding.
        TimeoutError
            Did not receive answer in time.
        TypeError
            - `proxy_authorization`'s type is incorrect.
            - ˙Cannot serialize a field of the given `data`.
        
        See Also
        --------
        - ``.request`` : Executes an http request returning a request context manager.
        - ``.request2`` : Executes an http request with extra parameters returning a request context manager.
        - ``._request2`` : Internal method for executing an http request with extra parameters.
        """
        if (params is not ...):
            warn(
                (
                    f'`{type(self).__name__}._request`\'s `params` parameter is deprecated '
                    f'and will be removed at 2026 January. '
                    'Please use `query` instead.'
                ),
                FutureWarning,
                stacklevel = 3,
            )
            
            query = params
        
        
        headers = IgnoreCaseMultiValueDictionary(headers)
        
        history = []
        url = URL(url)
        
        with Timeout(self.loop, REQUEST_TIMEOUT_DEFAULT):
            while True:
                cookies = self.cookie_jar.filter_cookies(url)
                
                request = ClientRequest(
                    self.loop,
                    method,
                    url,
                    headers,
                    data,
                    query,
                    cookies,
                    None,
                    None,
                    self.proxy,
                    None,
                    None,
                )
                
                connection = await self.connector.connect(request)
                
                set_tcp_nodelay(connection.get_transport(), True)
                
                response = request.begin(connection)
                await response.start_processing()
                
                # we do nothing with os error
                
                self.cookie_jar.update_cookies(response.cookies, response.url)
                
                # redirects
                if response.status in (301, 302, 303, 307) and redirects:
                    redirects -= 1
                    history.append(response)
                    if not redirects:
                        response.close()
                        raise ConnectionError('Too many redirects', history[0].request_info, tuple(history))
                    
                    if (
                        (response.status == 303 and response.method != METHOD_HEAD)
                        or (response.status in (301, 302) and response.method == METHOD_POST)
                    ):
                        method = METHOD_GET
                        data = None
                        try:
                            del headers[CONTENT_LENGTH]
                        except KeyError:
                            pass
                    
                    redirect_url = response.headers.get(LOCATION, None)
                    if redirect_url is None:
                        redirect_url = response.headers.get(URI, None)
                        if redirect_url is None:
                            break
                    
                    response.release()
                    
                    redirect_url = URL(redirect_url)
                    
                    scheme = redirect_url.scheme
                    if scheme not in ('http', 'https', ''):
                        response.close()
                        raise ConnectionError(
                            f'Can redirect only to http or https, got {scheme!r}',
                            history[0].request_info,
                            tuple(history),
                        )
                    
                    elif not scheme:
                        redirect_url = url.join(redirect_url)
                    
                    if url.origin() != redirect_url.origin():
                        try:
                            del headers[AUTHORIZATION]
                        except KeyError:
                            pass
                    
                    url = redirect_url
                    query = None
                    response.release()
                    continue
                
                break
        
        response.history = tuple(history)
        return response
    
    
    async def _request2(
        self,
        method,
        url,
        headers,
        *,
        authorization = None,
        data = None,
        params = ...,
        proxy = ...,
        query = None,
        redirects = 3,
        timeout = REQUEST_TIMEOUT_DEFAULT,
        ssl = ...,
        ssl_context = None,
        ssl_fingerprint = None,
        # Deprecated ones:
        auth = ...,
        proxy_auth = ...,
        proxy_url = ...,
        proxy_headers = ...,
    ):
        """
        Internal method for executing an http request with extra parameters
        
        This method is a coroutine.
        
        Parameters
        ----------
        method : `str`
            The method of the request.
        
        url : `str`, ``URL``
            The url to request.
        
        headers : `None | dict<str, str> | IgnoreCaseMultiValueDictionary`
            Request headers.
        
        authorization : `None`, ``BasicAuthorization`` = `None`, Optional (Keyword only)
            Authorization to use.
        
        data : `None`, `object` = `None`, Optional (Keyword only)
            Data to send a the body of the request.
        
        proxy : `None | Proxy`, Optional
            Proxy to use with the request.
        
        query : `None | str | dict<None, str | bool | int | float, iterable<...>> | iterable<...>` = `None`, \
                Optional (Keyword only)
        
        redirects : `int` = `3`, Optional (Keyword only)
            The maximal amount of allowed redirects.
        
        timeout : `float` = `REQUEST_TIMEOUT_DEFAULT`, Optional (Keyword only)
            The maximal duration to wait for server response. Defaults to `60.0` seconds.
        
        ssl_context : `None | SSLContext`, Optional (Keyword only)
            SSL context to be used by the connector.
        
        ssl_fingerprint : `None | SSLFingerprint`, Optional (Keyword only)
            SSL finger print to use by the connector.
        
        Returns
        -------
        response : ``ClientResponse``
        
        Raises
        ------
        ConnectionError
            - Too many redirects.
            - Connector closed.
        TypeError
            - `proxy_authorization`'s type is incorrect.
            - ˙Cannot serialize a field of the given `data`.
        ValueError
            - Host could not be detected from `url`.
            - `compression` and `Content-Encoding` would be set at the same time.
            - `chunked` cannot be set, because `Transfer-Encoding: chunked` is already set.
            - `chunked` cannot be set, because `Content-Length` header is already present.
            - `headers` contain authorization headers, but `authorization` parameter is given as well.
        RuntimeError
            - If one of `data`'s field's content has unknown content-encoding.
            - If one of `data`'s field's content has unknown content-transfer-encoding.
        TimeoutError
            - Did not receive answer in time.
        
        See Also
        --------
        - ``.request`` : Executes an http request returning a request context manager.
        - ``.request2`` : Executes an http request with extra parameters returning a request context manager.
        - ``._request`` : Internal method for executing an http request without extra parameters.
        """
        # Deprecations
        if (ssl is not ...):
            warn(
                (
                    f'`{type(self).__name__}._request2`\'s `ssl` parameter is deprecated '
                    f'and scheduled for removal in 2025 August. '
                    f'Please use either the `ssl_context` or the `ssl_fingerprint` parameters depending on your needs.'
                ),
                FutureWarning,
                stacklevel = 3,
            )
            
            if ssl is None:
                ssl_context = None
                ssl_fingerprint = None
            
            elif isinstance(ssl, SSLContext):
                ssl_context = ssl
                ssl_fingerprint = None
            
            elif isinstance(ssl, SSLFingerprint):
                ssl_context = None
                ssl_fingerprint = ssl
            
            elif isinstance(ssl, bool):
                ssl_context = create_default_ssl_context()
                ssl_fingerprint = None
                
            else:
                raise TypeError(
                    f'`ssl` can be `None`, `bool`, `{SSLContext.__name__}`, `{SSLFingerprint.__name__}`. '
                    f'Got {type(ssl).__name__}; {ssl!r}.'
                )
        
        if auth is not ...:
            warn(
                (
                    f'`{type(self).__name__}._request2`\'s `auth` parameters has been deprecated and will be '
                    f'removed at 2025 November. Please use `authorization` instead.'
                ),
                FutureWarning,
                stacklevel = 3,
            )
            authorization = auth
        
        
        if (proxy_auth is not ...):
            warn(
                (
                    f'`{type(self).__name__}._request2`\'s `proxy_auth` is deprecated '
                    f'and will be removed at 2025 November. '
                    'Please use `proxy = Proxy(url, authorization = authorization)` instead.'
                ),
                FutureWarning,
                stacklevel = 3,
            )
        
        if (proxy_headers is not ...):
            warn(
                (
                    f'`{type(self).__name__}._request2`\'s `proxy_headers` is deprecated '
                    f'and will be removed at 2025 November. '
                    'Please use `proxy = Proxy(url, headers = headers)` instead.'
                ),
                FutureWarning,
                stacklevel = 3,
            )
        
        if (proxy_url is not ...):
            warn(
                (
                    f'`{type(self).__name__}._request2`\'s `proxy_url` is deprecated '
                    f'and will be removed at 2025 November. '
                    'Please use `proxy = Proxy(url)` instead.'
                ),
                FutureWarning,
                stacklevel = 3,
            )
        
        if (params is not ...):
            warn(
                (
                    f'`{type(self).__name__}._request2`\'s `params` parameter is deprecated '
                    f'and will be removed at 2026 January. '
                    'Please use `query` instead.'
                ),
                FutureWarning,
                stacklevel = 3,
            )
            
            query = params
        
        # Transform headers to IgnoreCaseMultiValueDictionary
        headers = IgnoreCaseMultiValueDictionary(headers)
        
        if (headers and (authorization is not None) and AUTHORIZATION in headers):
            raise ValueError(
                f'Can\'t combine {AUTHORIZATION!r} header with the `authorization` parameter. '
                f'Got authorization = {authorization!r}; headers[{AUTHORIZATION!r}] = {headers[AUTHORIZATION]!r}.'
            )
        
        # proxy
        if proxy is ...:
            if (proxy_url is ...):
                proxy = self.proxy
            else:
                proxy = Proxy(proxy_url, headers = proxy_headers, authorization = proxy_auth)
        
        elif (proxy is not None) and (not isinstance(proxy, Proxy)):
            raise TypeError(
                f'`proxy` can be `None`, `{Proxy.__name__}`, got '
                f'{type(proxy).__name__}; {proxy!r}.'
            )
        
        history = []
        url = URL(url)
        
        with Timeout(self.loop, timeout):
            while True:
                cookies = self.cookie_jar.filter_cookies(url)
                
                request = ClientRequest(
                    self.loop,
                    method,
                    url,
                    headers,
                    data,
                    query,
                    cookies,
                    authorization,
                    None,
                    proxy,
                    ssl_context,
                    ssl_fingerprint,
                )
                
                connection = await self.connector.connect(request)
                
                set_tcp_nodelay(connection.get_transport(), True)
                
                response = request.begin(connection)
                await response.start_processing()
                
                # we do nothing with os error
                
                self.cookie_jar.update_cookies(response.cookies, response.url)
                
                # redirects
                if response.status in (301, 302, 303, 307) and redirects:
                    redirects -= 1
                    history.append(response)
                    if not redirects:
                        response.close()
                        raise ConnectionError('Too many redirects', history[0].request_info, tuple(history))
                    
                    # For 301 and 302, mimic IE behaviour, now changed in RFC.
                    if (
                        (response.status == 303 and response.method != METHOD_HEAD)
                        or (response.status in (301, 302) and response.method == METHOD_POST)
                    ):
                        method = METHOD_GET
                        data = None
                        content_length = headers.get(CONTENT_LENGTH, None)
                        if (content_length is not None) and content_length:
                            del headers[CONTENT_LENGTH]
                    
                    redirect_url = response.headers.get(LOCATION, None)
                    if redirect_url is None:
                        redirect_url = response.headers.get(URI, None)
                        if redirect_url is None:
                            break
                    
                    response.release()
                    
                    redirect_url = URL(redirect_url)
                    
                    scheme = redirect_url.scheme
                    if scheme not in ('http', 'https', ''):
                        response.close()
                        raise ConnectionError(
                            f'Can redirect only to http or https, got {scheme!r}',
                            history[0].request_info,
                            tuple(history),
                        )
                    
                    elif not scheme:
                        redirect_url = url.join(redirect_url)
                    
                    url = redirect_url
                    query = None
                    continue
                
                break
        
        response.history = tuple(history)
        return response
    
    
    @property
    def closed(self):
        """
        Returns whether the ``HTTPClient`` is closed.
        
        Returns
        -------
        closed : `bool`
        """
        connector = self.connector
        if connector is None:
            return True
        
        if connector.closed:
            return True
        
        return False
    
    
    async def __aenter__(self):
        """
        Enters the http client as an asynchronous context manager.
        
        This method is a coroutine.
        """
        return self
    
    
    async def __aexit__(self, exception_type, exception_value, exception_traceback):
        """
        Exits the http client with closing it.
        
        This method is a coroutine.
        """
        self.close()
    
    
    def __del__(self):
        """
        Closes the http client closed.
        """
        connector = self.connector
        if connector is None:
            return
        
        self.connector = None
        
        if not connector.closed:
            connector.close()
    
    
    close = __del__
    
    
    def request(self, method, url, headers = None, **keyword_parameters):
        """
        Executes an http request.
        
        Parameters
        ----------
        method : `str`
            The method of the request.
        
        url : `str | URL`
            The url to request.
        
        headers : `None | dict<str, str> | IgnoreCaseMultiValueDictionary` = `None`, Optional
            Request headers.
        
        **keyword_parameters : Keyword Parameters
            Additional keyword parameters.
        
        Other Parameters
        ----------------
        data : `None | object`, Optional (Keyword only)
            Data to send a the body of the request.
        
        query : `None | str | dict<None, str | bool | int | float, iterable<...>> | iterable<...>`, \
                Optional (Keyword only)
            Query string parameters.
        
        redirects : `int`, Optional (Keyword only)
            The maximal amount of allowed redirects.
        
        Returns
        -------
        request_context_manager : ``RequestContextManager``
        
        See Also
        --------
        - ``.request2`` : Executes an http request with extra parameters returning a request context manager.
        - ``.get`` : Shortcut for executing a get request.
        - ``.options`` : Shortcut for executing an options request.
        - ``.head`` : Shortcut for executing a head request.
        - ``.post`` : Shortcut for executing a post request.
        - ``.put`` : Shortcut for executing a put request.
        - ``.patch`` : Shortcut for executing a patch request.
        - ``.delete`` :  Shortcut for executing a delete request.
        """
        return RequestContextManager(self._request(method, url, headers, **keyword_parameters))
    
    
    def request2(self, method, url, headers = None, **keyword_parameters):
        """
        Executes an http request with extra parameters.
        
        Parameters
        ----------
        method : `str`
            The method of the request.
        
        url : `str | URL`
            The url to request.
        
        headers : `None | dict<str, str> | IgnoreCaseMultiValueDictionary` = `None`, Optional
            Request headers.
        
        **keyword_parameters : Keyword Parameters
            Additional keyword parameters.
        
        Other Parameters
        ----------------
        authorization : `None | BasicAuthorization`, Optional (Keyword only)
            Authorization to use.
        
        data : `None | object`, Optional (Keyword only)
            Data to send a the body of the request.
        
        query : `None | str | dict<None, str | bool | int | float, iterable<...>> | iterable<...>`, \
                Optional (Keyword only)
            Query string parameters.
        
        redirects : `int`, Optional (Keyword only)
            The maximal amount of allowed redirects.
        
        proxy_authorization : `None | BasicAuthorization`, Optional (Keyword only)
            Proxy authorization to use instead of the client's.
        
        proxy_headers : `None | dict<str, str> | IgnoreCaseMultiValueDictionary`, Optional (Keyword only)
            Proxy headers to use instead of the client's.
        
        proxy_url : `None | str | URL`, Optional  (Keyword only)
            Proxy url to use instead of the client's own.
        
        ssl_context : `None | SSLContext`, Optional (Keyword only)
            SSL context to be used by the connector.
        
        ssl_fingerprint : `None | SSLFingerprint`, Optional (Keyword only)
            SSL finger print to use by the connector.
        
        timeout : `float`, Optional (Keyword only)
            The maximal duration to wait for server response.
        
        Returns
        -------
        request_context_manager : ``RequestContextManager``
        
        See Also
        --------
        - ``.request`` : Executes an http request without extra parameters returning a request context manager.
        - ``.get`` : Shortcut for executing a get request.
        - ``.options`` : Shortcut for executing an options request.
        - ``.head`` : Shortcut for executing a head request.
        - ``.post`` : Shortcut for executing a post request.
        - ``.put`` : Shortcut for executing a put request.
        - ``.patch`` : Shortcut for executing a patch request.
        - ``.delete`` :  Shortcut for executing a delete request.
        """
        return RequestContextManager(self._request2(method, url, headers, **keyword_parameters))
    
    
    def get(self, url, headers = None, **keyword_parameters):
        """
        Shortcut for executing a get request.
        
        Parameters
        ----------
        url : `str | URL`
            The url to request.
        
        headers : `None | dict<str, str> | IgnoreCaseMultiValueDictionary` = `None`, Optional
            Request headers.
        
        **keyword_parameters : Keyword Parameters
            Additional keyword parameters.
        
        Other Parameters
        ----------------
        data : `None | object`, Optional (Keyword only)
            Data to send a the body of the request.
        
        query : `None | str | dict<None, str | bool | int | float, iterable<...>> | iterable<...>`, \
                Optional (Keyword only)
            Query string parameters.
        
        redirects : `int`, Optional (Keyword only)
            The maximal amount of allowed redirects.

        Returns
        -------
        request_context_manager : ``RequestContextManager``
        
        See Also
        --------
        - ``.request`` : Executes an http request without extra parameters returning a request context manager.
        - ``.request2`` : Executes an http request with extra parameters returning a request context manager.
        - ``.options`` : Shortcut for executing an options request.
        - ``.head`` : Shortcut for executing a head request.
        - ``.post`` : Shortcut for executing a post request.
        - ``.put`` : Shortcut for executing a put request.
        - ``.patch`` : Shortcut for executing a patch request.
        - ``.delete`` :  Shortcut for executing a delete request.
        """
        return RequestContextManager(self._request(METHOD_GET, url, headers, **keyword_parameters))
    
    
    def options(self, url, headers = None, **keyword_parameters):
        """
        Shortcut for executing a get request.
        
        Parameters
        ----------
        url : `str | URL`
            The url to request.
        
        headers : `None | dict<str, str> | IgnoreCaseMultiValueDictionary` = `None`, Optional
            Request headers.
        
        **keyword_parameters : Keyword Parameters
            Additional keyword parameters.
        
        Other Parameters
        ----------------
        data : `None | object`, Optional (Keyword only)
            Data to send a the body of the request.
        
        query : `None | str | dict<None, str | bool | int | float, iterable<...>> | iterable<...>`, \
                Optional (Keyword only)
            Query string parameters.
        
        redirects : `int`, Optional (Keyword only)
            The maximal amount of allowed redirects.

        Returns
        -------
        request_context_manager : ``RequestContextManager``
        
        See Also
        --------
        - ``.request`` : Executes an http request without extra parameters returning a request context manager.
        - ``.request2`` : Executes an http request with extra parameters returning a request context manager.
        - ``.get`` : Shortcut for executing a get request.
        - ``.head`` : Shortcut for executing a head request.
        - ``.post`` : Shortcut for executing a post request.
        - ``.put`` : Shortcut for executing a put request.
        - ``.patch`` : Shortcut for executing a patch request.
        - ``.delete`` :  Shortcut for executing a delete request.
        """
        return RequestContextManager(self._request(METHOD_OPTIONS, url, headers, **keyword_parameters))
    
    
    def head(self, url, headers = None, **keyword_parameters):
        """
        Shortcut for executing a head request.
        
        Parameters
        ----------
        url : `str | URL`
            The url to request.
        
        headers : `None | dict<str, str> | IgnoreCaseMultiValueDictionary` = `None`, Optional
            Request headers.
        
        **keyword_parameters : Keyword Parameters
            Additional keyword parameters.
        
        Other Parameters
        ----------------
        data : `None | object`, Optional (Keyword only)
            Data to send a the body of the request.
        
        query : `None | str | dict<None, str | bool | int | float, iterable<...>> | iterable<...>`, \
                Optional (Keyword only)
            Query string parameters.
        
        redirects : `int`, Optional (Keyword only)
            The maximal amount of allowed redirects.

        Returns
        -------
        request_context_manager : ``RequestContextManager``
        
        See Also
        --------
        - ``.request`` : Executes an http request without extra parameters returning a request context manager.
        - ``.request2`` : Executes an http request with extra parameters returning a request context manager.
        - ``.get`` : Shortcut for executing a get request.
        - ``.options`` : Shortcut for executing an options request.
        - ``.post`` : Shortcut for executing a post request.
        - ``.put`` : Shortcut for executing a put request.
        - ``.patch`` : Shortcut for executing a patch request.
        - ``.delete`` :  Shortcut for executing a delete request.
        """
        return RequestContextManager(self._request(METHOD_HEAD, url, headers, **keyword_parameters))
    
    
    def post(self, url, headers = None, **keyword_parameters):
        """
        Shortcut for executing a post request.
        
        Parameters
        ----------
        url : `str | URL`
            The url to request.
        
        headers : `None | dict<str, str> | IgnoreCaseMultiValueDictionary` = `None`, Optional
            Request headers.
        
        **keyword_parameters : Keyword Parameters
            Additional keyword parameters.
        
        Other Parameters
        ----------------
        data : `None | object`, Optional (Keyword only)
            Data to send a the body of the request.
        
        query : `None | str | dict<None, str | bool | int | float, iterable<...>> | iterable<...>`, \
                Optional (Keyword only)
            Query string parameters.
        
        redirects : `int`, Optional (Keyword only)
            The maximal amount of allowed redirects.

        Returns
        -------
        request_context_manager : ``RequestContextManager``
        
        See Also
        --------
        - ``.request`` : Executes an http request without extra parameters returning a request context manager.
        - ``.request2`` : Executes an http request with extra parameters returning a request context manager.
        - ``.get`` : Shortcut for executing a get request.
        - ``.options`` : Shortcut for executing an options request.
        - ``.head`` : Shortcut for executing a head request.
        - ``.put`` : Shortcut for executing a put request.
        - ``.patch`` : Shortcut for executing a patch request.
        - ``.delete`` :  Shortcut for executing a delete request.
        """
        return RequestContextManager(self._request(METHOD_POST, url, headers, **keyword_parameters))
    
    
    def put(self, url, headers = None, **keyword_parameters):
        """
        Shortcut for executing a put request.
        
        Parameters
        ----------
        url : `str | URL`
            The url to request.
        
        headers : `None | dict<str, str> | IgnoreCaseMultiValueDictionary` = `None`, Optional
            Request headers.
        
        **keyword_parameters : Keyword Parameters
            Additional keyword parameters.
        
        Other Parameters
        ----------------
        data : `None | object`, Optional (Keyword only)
            Data to send a the body of the request.
        
        query : `None | str | dict<None, str | bool | int | float, iterable<...>> | iterable<...>`, \
                Optional (Keyword only)
            Query string parameters.
        
        redirects : `int`, Optional (Keyword only)
            The maximal amount of allowed redirects.

        Returns
        -------
        request_context_manager : ``RequestContextManager``
        
        See Also
        --------
        - ``.request`` : Executes an http request without extra parameters returning a request context manager.
        - ``.request2`` : Executes an http request with extra parameters returning a request context manager.
        - ``.get`` : Shortcut for executing a get request.
        - ``.options`` : Shortcut for executing an options request.
        - ``.head`` : Shortcut for executing a head request.
        - ``.post`` : Shortcut for executing a post request.
        - ``.patch`` : Shortcut for executing a patch request.
        - ``.delete`` :  Shortcut for executing a delete request.
        """
        return RequestContextManager(self._request(METHOD_PUT, url, headers, **keyword_parameters))
    
    
    def patch(self, url, headers = None, **keyword_parameters):
        """
        Shortcut for executing a patch request.
        
        Parameters
        ----------
        url : `str | URL`
            The url to request.
        
        headers : `None | dict<str, str> | IgnoreCaseMultiValueDictionary` = `None`, Optional
            Request headers.
        
        **keyword_parameters : Keyword Parameters
            Additional keyword parameters.
        
        Other Parameters
        ----------------
        data : `None | object`, Optional (Keyword only)
            Data to send a the body of the request.
        
        query : `None | str | dict<None, str | bool | int | float, iterable<...>> | iterable<...>`, \
                Optional (Keyword only)
            Query string parameters.
        
        redirects : `int`, Optional (Keyword only)
            The maximal amount of allowed redirects.

        Returns
        -------
        request_context_manager : ``RequestContextManager``
        
        See Also
        --------
        - ``.request`` : Executes an http request without extra parameters returning a request context manager.
        - ``.request2`` : Executes an http request with extra parameters returning a request context manager.
        - ``.get`` : Shortcut for executing a get request.
        - ``.options`` : Shortcut for executing an options request.
        - ``.head`` : Shortcut for executing a head request.
        - ``.post`` : Shortcut for executing a post request.
        - ``.put`` : Shortcut for executing a put request.
        - ``.delete`` :  Shortcut for executing a delete request.
        """
        return RequestContextManager(self._request(METHOD_PATCH, url, headers, **keyword_parameters))
    
    
    def delete(self, url, headers = None, **keyword_parameters):
        """
        Shortcut for executing a delete request.
        
        Parameters
        ----------
        url : `str | URL`
            The url to request.
        
        headers : `None | dict<str, str> | IgnoreCaseMultiValueDictionary` = `None`, Optional
            Request headers.
        
        **keyword_parameters : Keyword Parameters
            Additional keyword parameters.
        
        Other Parameters
        ----------------
        data : `None | object`, Optional (Keyword only)
            Data to send a the body of the request.
        
        query : `None | str | dict<None, str | bool | int | float, iterable<...>> | iterable<...>`, \
                Optional (Keyword only)
            Query string parameters.
        
        redirects : `int`, Optional (Keyword only)
            The maximal amount of allowed redirects.

        Returns
        -------
        request_context_manager : ``RequestContextManager``
        
        See Also
        --------
        - ``.request`` : Executes an http request without extra parameters returning a request context manager.
        - ``.request2`` : Executes an http request with extra parameters returning a request context manager.
        - ``.get`` : Shortcut for executing a get request.
        - ``.options`` : Shortcut for executing an options request.
        - ``.head`` : Shortcut for executing a head request.
        - ``.post`` : Shortcut for executing a post request.
        - ``.put`` : Shortcut for executing a put request.
        - ``.patch`` : Shortcut for executing a patch request.
        """
        return RequestContextManager(self._request(METHOD_DELETE, url, headers, **keyword_parameters))
    
    
    def connect_websocket(self, url, **keyword_parameters):
        """
        Deprecated and will be removed in 2025 august. please use ``.connect_web_socket``.
        """
        warn(
            (
                f'`{type(self).__name__}.connect_websocket` is deprecated and will be removed in 2025 august. '
                f'Please use `.connect_web_socket` instead.'
            ),
            FutureWarning,
            stacklevel = 3,
        )
        return self.connect_web_socket(url, **keyword_parameters)
    
    
    def connect_web_socket(self, url, **keyword_parameters):
        """
        Connect a web socket client to the given url.
        
        Parameters
        ----------
        url : `str | URL`
            The url to connect to.
        
        **keyword_parameters : Keyword Parameters
            Additional keyword parameters.
        
        Other Parameters
        ----------------
        origin : `None`, `str`, Optional (Keyword only)
            Value of the Origin header.
        
        available_extensions : `None` or (`list` of `object`), Optional (Keyword only)
            Available web socket extensions.
            
            Each web socket extension should have the following `4` attributes / methods:
            - `name`: `str`. The extension's name.
            - `request_params` : `list` of `tuple` (`str`, `str`). Additional header parameters of the extension.
            - `decode` : `callable`. Decoder method, what processes a received web socket frame. Should accept `2`
                parameters: The respective ``WebSocketFrame``, and the ˙max_size` as `int`, what describes the
                maximal size of a received frame. If it is passed, ``PayloadError`` is raised.
            - `encode` : `callable`. Encoder method, what processes the web socket frames to send. Should accept `1`
                parameter, the respective ``WebSocketFrame``.
        
        available_subprotocols : `None | list<str>`, Optional (Keyword only)
            A list of supported subprotocols in order of decreasing preference.
        
        headers : `None | dict<str, str> | IgnoreCaseMultiValueDictionary` = `None`, Optional
            Request headers.
        
        http_client : `None | HTTPClient`, Optional (Keyword only)
            Http client to use to connect the web socket.
        
        close_timeout : `float`, Optional (Keyword only)
            The maximal duration in seconds what is waited for response after close frame is sent.
            Defaults to `10.0`.
        
        max_size : `int`, Optional (Keyword only)
            Max payload size to receive. If a payload exceeds it, ``PayloadError`` is raised.
            Defaults to `67108864` bytes.
        
        max_queue : `None`, `int`, Optional (Keyword only)
            Max queue size of ``.messages``.
            If a new payload is added to a full queue, the oldest element of it is removed.
        
        Returns
        -------
        web_socket_context_manager : ``WebSocketContextManager``
        """
        return WebSocketContextManager(WebSocketClient(self.loop, url, **keyword_parameters, http_client = self))



__all__ = ('WebSocketContextManager', )

from ..utils import RichAttributeErrorBaseType
from ..utils.trace.formatters import format_coroutine


class WebSocketContextManager(RichAttributeErrorBaseType):
    """
    Asynchronous context manager wrapping a web socket connecting coroutine.
    
    Examples
    --------
    web socket context managers are returned by the ``HTTPClient.connect_web_socket`` method. Web socket
    context managers can be used as an asynchronous context manager or as a simple awaitable.
    
    ```py
    async with http_client.connect_web_socket('http://ayaya.aya') as web_socket:
        await web_socket.send('ayaya')
    ```
    
    ```py
    web_socket = await http_client.connect_web_socket('http://ayaya.aya')
    await web_socket.send('ayaya')
    ```
    
    Attributes
    ----------
    coroutine : `None | CoroutineType`
        The wrapped requester coroutine.
    
    web_socket : `None | WebSocketClient`
        The connected web socket client if applicable
    """
    __slots__ = ('coroutine', 'web_socket')
    
    def __new__(cls, coroutine):
        """
        Creates a new web socket content manager.
        
        Parameters
        ----------
        coroutine : `CoroutineType`
            WebSocket connecting coroutine to wrap.
        """
        self = object.__new__(cls)
        self.coroutine = coroutine
        self.web_socket = None
        return self
    
    
    def __iter__(self):
        """
        Awaits the wrapped coroutine.
        
        This method is a generator. Should be used with `await` expression.
        
        Returns
        -------
        web_socket : ``WebSocketClient``
            The connected web socket client.
        
        Raises
        ------
        BaseException
            Any exception raised meanwhile connecting the web socket.
        """
        return (yield from self.__aenter__().__await__())
    
    __await__ = __iter__
    
    
    async def __aenter__(self):
        """
        Enters the web socket context manager as an asynchronous context manager. Closes the web socket when the
        context manager is exited.
        
        This method is a coroutine.
        
        Returns
        -------
        web_socket : ``WebSocketClient``
            The connected web socket client.
        
        Raises
        ------
        BaseException
            Any exception raised meanwhile connecting the web socket.
        """
        coroutine = self.coroutine
        if (coroutine is None):
            web_socket = self.web_socket
        else:
            web_socket = await coroutine
            self.web_socket = web_socket
            self.coroutine = None
        
        return web_socket
    
    
    async def __aexit__(self, exception_type, exception_value, exception_traceback):
        """
        Closes the connected web_socket.
        
        This method is a coroutine.
        """
        web_socket = self.web_socket
        if (web_socket is not None):
            self.web_socket = None
            await web_socket.close()
    
    
    def __repr__(self):
        """Returns the request context manager's representation."""
        repr_parts = ['<', type(self).__name__]
        
        while True:
            coroutine = self.coroutine
            if (coroutine is not None):
                repr_parts.append(' coroutine = ')
                repr_parts.append(format_coroutine(coroutine))
                break
            
            web_socket = self.web_socket
            if (web_socket is not None):
                repr_parts.append(' web_socket = ')
                repr_parts.append(repr(web_socket))
                break
            
            repr_parts.append(' exited')
            break
        
        repr_parts.append('>')
        return ''.join(repr_parts)



__all__ = ('ExpressionKey',)


class ExpressionKey:
    """
    Key used to represent an expression.
    
    Attributes
    ----------
    file_name : `str`
        The file name the expression is located at.
    instruction_index : `int`
        The instruction's index used for identification.
    line_index : `int`
        The line's identifier where the expression is at.
    name : `str`
        The function's name the expression is at.
    """
    __slots__ = ('file_name', 'instruction_index', 'line_index', 'name')
    
    def __new__(cls, file_name, line_index, name, instruction_index):
        """
        Creates a new expression key.
        
        Parameters
        ----------
        file_name : `str`
            The file name the expression is located at.
        line_index : `int`
            The line's identifier where the expression is at.
        name : `str`
            The function's name the expression is at.
        instruction_index : `int`
            The instruction's index used for identification.
        """
        self = object.__new__(cls)
        self.file_name = file_name
        self.instruction_index = instruction_index
        self.line_index = line_index
        self.name = name
        return self
    
    
    def __repr__(self):
        """Returns the expression key's representation."""
        repr_parts = ['<', type(self).__name__]
        
        repr_parts.append(' file_name = ')
        repr_parts.append(repr(self.file_name))
        
        repr_parts.append(', line_index = ')
        repr_parts.append(repr(self.line_index))
        
        repr_parts.append(', name = ')
        repr_parts.append(repr(self.name))
        
        repr_parts.append(', instruction_index = ')
        repr_parts.append(repr(self.instruction_index))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two expression keys are equal."""
        if type(self) is not type(other):
            return False
        
        if self.file_name != other.file_name:
            return False
        
        if self.instruction_index != other.instruction_index:
            return False
        
        if self.line_index != other.line_index:
            return False
        
        if self.name != other.name:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the expression key's hash value."""
        hash_value = 0
        hash_value ^= hash(self.file_name)
        hash_value ^= self.instruction_index
        hash_value ^= self.line_index << 16
        hash_value ^= hash(self.name)
        return hash_value



__all__ = ('ExceptionRepresentationAttributeError',)

from ...docs import copy_docs

from .attribute_error_helpers import extract_attribute_error_fields, extract_last_attribute_error_context_frame
from .exception_representation_base import ExceptionRepresentationBase
from .representation_helpers import _get_type_name
from .suggestion import (
    exists_matching_variable_name, get_familiar_instance_attribute_names, get_matching_variable_names_with_attribute
)


class ExceptionRepresentationAttributeError(ExceptionRepresentationBase):
    """
    Holds a valid syntax error's representation.
    
    Attributes
    ----------
    attribute_name : `str`
        The attribute's name that was not present.
    instance_type_name : `str`
        The representation of the instance's type's name.
    suggestion_attribute_unset : `bool`
        Whether the instance should have an attribute with the given name.
    suggestion_familiar_attribute_names : `None | list<str>`
        Attributes of the same instance with familiar names.
    suggestion_matching_variable_exists : `bool`
        Variables of the same scope with the same name.
    suggestion_variable_names_with_attribute : `None | list<str>`
        Variables with an attribute with the same name.
    type_name : `str`
        The represented exception's type's name.
    """
    __slots__ = (
        'attribute_name', 'instance_type_name', 'suggestion_attribute_unset',
        'suggestion_familiar_attribute_names', 'suggestion_matching_variable_exists', 'suggestion_matching_variables',
        'suggestion_variable_names_with_attribute', 'type_name',
    )
    
    def __new__(cls, exception, frames):
        """
        Creates a new exception representation.
        
        Parameters
        ----------
        exception : `AttributeError`
            Exception to represent.
        frames : `None | list<FrameProxyBase>`
            The frame the exception is raised with.
        """
        instance, attribute_name = extract_attribute_error_fields(exception)
        frame = extract_last_attribute_error_context_frame(exception, frames)
        
        suggestion_attribute_unset, suggestion_familiar_attribute_names = (
            get_familiar_instance_attribute_names(instance, attribute_name)
        )
        suggestion_matching_variable_exists = exists_matching_variable_name(frame, attribute_name)
        suggestion_variable_names_with_attribute = get_matching_variable_names_with_attribute(frame, attribute_name)
        
        instance_type_name = _get_type_name(type(instance))
        type_name = _get_type_name(type(exception))
        
        self = object.__new__(cls)
        self.attribute_name = attribute_name
        self.instance_type_name = instance_type_name
        self.suggestion_attribute_unset = suggestion_attribute_unset
        self.suggestion_familiar_attribute_names = suggestion_familiar_attribute_names
        self.suggestion_matching_variable_exists = suggestion_matching_variable_exists
        self.suggestion_variable_names_with_attribute = suggestion_variable_names_with_attribute
        self.type_name = type_name
        return self
    
    
    @classmethod
    def from_fields(
        cls,
        *,
        attribute_name = ...,
        instance_type_name = ...,
        suggestion_attribute_unset = ...,
        suggestion_familiar_attribute_names = ...,
        suggestion_matching_variable_exists = ...,
        suggestion_variable_names_with_attribute = ...,
        type_name = ...,
    ):
        """
        Creates a new syntax error representation from the given fields.
        
        Attributes
        ----------
        attribute_name : `str`, Optional (Keyword only)
            The attribute's name that was not present.
        instance_type_name : `str`, Optional (Keyword only)
            The representation of the instance's type's name.
        suggestion_attribute_unset : `bool`, Optional (Keyword only)
            Whether the instance should have an attribute with the given name.
        suggestion_familiar_attribute_names : `None | list<str>`, Optional (Keyword only)
            Attributes of the same instance with familiar names.
        suggestion_matching_variable_exists : `bool`, Optional (Keyword only)
            Variables of the same scope with the same name.
        suggestion_variable_names_with_attribute : `None | list<str>`, Optional (Keyword only)
            Variables with an attribute with the same name.
        type_name : `str`, Optional (Keyword only)
            The represented exception's type's name.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.attribute_name = '' if attribute_name is ... else attribute_name
        self.instance_type_name = '' if instance_type_name is ... else instance_type_name
        self.suggestion_attribute_unset = (
            False if suggestion_attribute_unset is ... else suggestion_attribute_unset
        )
        self.suggestion_familiar_attribute_names = (
            None if suggestion_familiar_attribute_names is ... else suggestion_familiar_attribute_names
        )
        self.suggestion_matching_variable_exists = (
            False if suggestion_matching_variable_exists is ... else suggestion_matching_variable_exists
        )
        self.suggestion_variable_names_with_attribute = (
            None if suggestion_variable_names_with_attribute is ... else suggestion_variable_names_with_attribute
        )
        self.type_name = AttributeError.__name__ if type_name is ... else type_name
        return self
    
    
    @copy_docs(ExceptionRepresentationBase._populate_repr_parts)
    def _populate_repr_parts(self, repr_parts):
        # type_name
        repr_parts.append(' type_name = ')
        repr_parts.append(repr(self.type_name))
        
        # instance_type_name
        repr_parts.append(', instance_type_name = ')
        repr_parts.append(repr(self.instance_type_name))
        
        # attribute_name
        repr_parts.append(', attribute_name = ')
        repr_parts.append(repr(self.attribute_name))
        
        # suggestion_attribute_unset
        suggestion_attribute_unset = self.suggestion_attribute_unset
        if suggestion_attribute_unset:
            repr_parts.append(', suggestion_attribute_unset = ')
            repr_parts.append(repr(suggestion_attribute_unset))
        
        # suggestion_familiar_attribute_names
        suggestion_familiar_attribute_names = self.suggestion_familiar_attribute_names
        if (suggestion_familiar_attribute_names is not None):
            repr_parts.append(', suggestion_familiar_attribute_names = ')
            repr_parts.append(repr(suggestion_familiar_attribute_names))
        
        # suggestion_matching_variable_exists
        suggestion_matching_variable_exists = self.suggestion_matching_variable_exists
        if suggestion_matching_variable_exists:
            repr_parts.append(', suggestion_matching_variable_exists = ')
            repr_parts.append(repr(suggestion_matching_variable_exists))
        
        # suggestion_variable_names_with_attribute
        suggestion_variable_names_with_attribute = self.suggestion_variable_names_with_attribute
        if (suggestion_variable_names_with_attribute is not None):
            repr_parts.append(', suggestion_variable_names_with_attribute = ')
            repr_parts.append(repr(suggestion_variable_names_with_attribute))
    
    
    @copy_docs(ExceptionRepresentationBase._is_equal)
    def _is_equal(self, other):
        if self.type_name != other.type_name:
            return False
        
        if self.attribute_name != other.attribute_name:
            return False
        
        if self.instance_type_name != other.instance_type_name:
            return False
        
        if self.suggestion_attribute_unset != other.suggestion_attribute_unset:
            return False
        
        if self.suggestion_familiar_attribute_names != other.suggestion_familiar_attribute_names:
            return False
        
        if self.suggestion_matching_variable_exists != other.suggestion_matching_variable_exists:
            return False
        
        if self.suggestion_variable_names_with_attribute != other.suggestion_variable_names_with_attribute:
            return False
        
        return True



__all__ = ()

from hashlib import md5, sha1, sha256
from re import compile as re_compile
from ssl import (
    OP_NO_COMPRESSION as SSL_OPTION_NO_COMPRESSION, OP_NO_SSLv2 as SSL_OPTION_NO_SSL_V2,
    OP_NO_SSLv3 as SSL_OPTION_SSL_NO_SSL_V3, PROTOCOL_SSLv23 as SSL_PROTOCOL_TLS, SSLContext
)

from ..web_common.headers import ACCEPT, ACCEPT_ENCODING


REQUEST_TIMEOUT_DEFAULT = 60.0
HOST_INFO_CACHE_TIMEOUT = 10.0
CONNECTION_KEEP_ALIVE_TIMEOUT = 15.0


DEFAULT_HEADERS = (
    (ACCEPT, '*/*'),
    (ACCEPT_ENCODING, 'gzip, deflate'),
)


HASH_FUNCTION_BY_DIGEST_LENGTH = {
    16: md5,
    20: sha1,
    32: sha256,
}


JSON_RE = re_compile(r'^application/(?:[\w.+-]+?\+)?json')


SSL_CONTEXT_UNVERIFIED = SSLContext(SSL_PROTOCOL_TLS)
SSL_CONTEXT_UNVERIFIED.options |= SSL_OPTION_NO_SSL_V2 | SSL_OPTION_SSL_NO_SSL_V3 | SSL_OPTION_NO_COMPRESSION
SSL_CONTEXT_UNVERIFIED.set_default_verify_paths()



__all__ = ('convert_frames_to_frame_proxies', 'get_exception_frames', 'populate_frame_proxies',)

from types import FrameType, TracebackType

from ..expression_parsing import ExpressionInfo, LineCacheSession, get_expression_info

from .frame_proxy_base import FrameProxyBase
from .frame_proxy_frame import FrameProxyFrame
from .frame_proxy_traceback import FrameProxyTraceback


def convert_frames_to_frame_proxies(frames):
    """
    Converts the given frames into frame proxies.
    
    Parameters
    ----------
    frames : `list<FrameProxyBase | FrameType | TracebackType>`
        The frames to convert.
    
    Returns
    -------
    frame_proxies : `list<FrameProxyBase>`
    
    Raises
    ------
    TypeError
        - Unknown frame type.
    """
    frame_proxies = []
    
    for frame in frames:
        if isinstance(frame, FrameProxyBase):
            frame_proxy = frame
        
        elif isinstance(frame, FrameType):
            frame_proxy = FrameProxyFrame(frame)
        
        elif isinstance(frame, TracebackType):
            frame_proxy = FrameProxyTraceback(frame)
        
        else:
            raise TypeError(
                f'Unknown frame type, got {type(frame).__name__}; {frame!r}; frames = {frames!r}.'
            )
        
        frame_proxies.append(frame_proxy)
    
    return frame_proxies


def get_exception_frames(exception):
    """
    Gets the frames of the given exception.
    
    Parameters
    ----------
    exception : `BaseException`
        The exception to trace back.
    
    Returns
    -------
    frames : `list<FrameProxyBase>`
        A list of `frame` compatible exception frames.
    """
    frames = []
    traceback = exception.__traceback__
    
    while True:
        if traceback is None:
            break
        
        frame = FrameProxyTraceback(traceback)
        frames.append(frame)
        traceback = traceback.tb_next
    
    return frames


def populate_frame_proxies(frames):
    """
    Populates the given frame proxies.
    
    Parameters
    ----------
    frames : `list<FrameProxyBase>`
        The frame proxies to populate.
    """
    expressions = {}
    
    for frame in frames:
        expression_key = frame.expression_key
        expression_info = frame.expression_info
        
        if expression_info is None:
            expressions.setdefault(expression_key, None)
        else:
            expressions[expression_key] = expression_info.copy()
    
    with LineCacheSession():
        for expression_key, expression_value in expressions.items():
            if expression_value is None:
                expressions[expression_key] = get_expression_info(expression_key)
    
    
    for frame in frames:
        expression_key = frame.expression_key
        
        try:
            expression_info = expressions[expression_key]
        except KeyError:
            # A frame changed while we were rendering, create a dummy 
            expression_info = ExpressionInfo(expression_key, [], 0, True)
            expressions.setdefault(expression_key, expression_info)
        
        frame.expression_info = expression_info



__all__ = ()

from itertools import islice, zip_longest
from os import get_terminal_size

from ....utils import DEFAULT_ANSI_HIGHLIGHTER, RichAttributeErrorBaseType, iter_highlight_code_lines

from .editor_base import _validate_buffer
from .line_render_intermediate import LineRenderIntermediate
from .terminal_control_commands import (
    COMMAND_CLEAR_LINE_FROM_CURSOR, COMMAND_CLEAR_LINE_WHOLE, COMMAND_FORMAT_RESET, COMMAND_DOWN, COMMAND_UP, COMMAND_START_LINE,
    create_command_down, create_command_left, create_command_right, create_command_up, is_command
)

EMPTY_LINE_PREFIX_CHARACTER = ' '
CONTINUOUS_LINE_POSTFIX = '\\'


def _get_line_line_count(line_length, content_width, new_content_width, prefix_length):
    """
    Returns how much lines the given line is displayed as.
    
    Parameters
    ----------
    line_length : `int`
        The line's length.
    content_width : `int`
        The content field's width.
    new_content_width : `int`
        The new column width to use.
    prefix_length : `int`
        Prefix length used for calculations when column width is changed.
    
    Returns
    -------
    line_line_count : `int`
    """
    if (content_width == -1) or (new_content_width >= content_width):
        return _get_line_line_count_same_or_increased(
            line_length,
            (new_content_width if content_width == -1 else content_width),
        )
    
    return _get_line_line_count_reduced(line_length, content_width, new_content_width, prefix_length)


def _get_line_line_count_same_or_increased(line_length, content_width):
    """
    Returns how much lines the given line is displayed as.
    
    This function handles the case when content width is not changed or is increased.
    Increase does not matter because when resizing the old linebreaks are kept, so the lines are not squashed.
    
    Parameters
    ----------
    line_length : `int`
        The line's length.
    content_width : `int`
        The content field's width.
    
    Returns
    -------
    line_line_count : `int`
    """
    if line_length == 0:
        line_line_count = 1
    else:
        line_line_count = 1 + (line_length - 1) // content_width
    
    return line_line_count


def _get_line_line_count_reduced(line_length, content_width, new_content_width, prefix_length):
    """
    Returns how much lines the given line is displayed as.
    
    This function handles the case when the content width is reduced and further calculations are required.
    
    Parameters
    ----------
    line_length : `int`
        The line's length.
    content_width : `int`
        The content field's width.
    new_content_width : `int`
        The new column width to use.
    prefix_length : `int`
        Prefix length used for calculations when column width is changed.
    
    Returns
    -------
    line_line_count : `int`
    """
    line_line_count = 0
    
    if line_length == 0:
        line_line_count += 1
    
    else:
        additional_length = prefix_length + len(CONTINUOUS_LINE_POSTFIX)
        old_full_line_length = content_width + additional_length
        new_full_line_length = new_content_width + additional_length
        
        full_line_count, last_line_length = divmod(line_length, content_width)
        
        # If the last line is full (so last_line_length == 0 and full_line_count > 0)
        # then that line does not have `/` at the end, so we should remove that
        # from `full_line_count` and add to `last_line_length` instead.
        if not last_line_length and full_line_count:
            full_line_count -= 1
            last_line_length = line_length
            
        if full_line_count:
            line_line_count += full_line_count * _get_line_line_count_same_or_increased(
                old_full_line_length, new_full_line_length
            )
        
        if last_line_length:
            line_line_count += 1 + ((last_line_length + prefix_length - 1) // new_full_line_length)
        
    return line_line_count


def _get_line_cursor_index(content_width, cursor_index, prefix_length):
    """
    Gets the cursor's index depending on the parameters.
    
    Parameters
    ----------
    content_width : `int`
        The content field's width.
    cursor_index : `int`
        The cursors index on the current line.
    prefix_length : `int`
        Prefix length used for calculations when column width is changed.
    
    Returns
    -------
    cursor_index : `int`
    """
    line_cursor_index = prefix_length
    if cursor_index:
        line_cursor_index += (cursor_index - 1) % content_width + 1
    
    return line_cursor_index


def _jump_vertical(write_buffer, old_index, new_index):
    """
    Jumps from the old position to the new position.
    
    Parameters
    ----------
    write_buffer : `list<str>`
        Write buffer to extend.
    old_index : `int`
        Old line index to jump to.
    new_index : `int`
        New line index to jump to.
    """
    if old_index > new_index:
        command = create_command_up(old_index - new_index)
    
    elif new_index > old_index:
        command = create_command_down(new_index - old_index)
    
    else:
        return
    
    write_buffer.append(command)


def _jump_horizontal(write_buffer, old_index, new_index):
    """
    Jumps from the old position to the new position.
    
    Parameters
    ----------
    write_buffer : `list<str>`
        Write buffer to extend.
    old_index : `int`
        Old line index to jump to.
    new_index : `int`
        New line index to jump to.
    """
    if old_index > new_index:
        command = create_command_left(old_index - new_index)
    
    elif new_index > old_index:
        command = create_command_right(new_index - old_index)
    
    else:
        return
    
    write_buffer.append(command)


def _cut_same_prefix(string_0, string_1):
    """
    Cuts the prefix of the two strings till they are same.
    
    Parameters
    ----------
    string_0 : `str`
        String to cut.
    string_1 : `str`
        String to cut.
    
    Returns
    -------
    cut_count : `int`
    cut_string_0 : `str`
    cut_string_1 : `str`
    """
    index = 0
    length_string_0 = len(string_0)
    length_string_1 = len(string_1)
    
    while index < length_string_0 and index < length_string_1:
        if string_0[index] != string_1[index]:
            break
        
        index += 1
        continue
    
    return index, string_0[index:], string_1[index:]


def get_new_content_width(prefix_length):
    """
    Gets the new column with and whether it changes.
    
    Parameters
    ----------
    prefix_length : `int`
        Prefix length used for calculations of content width.
    
    Returns
    -------
    new_content_width : `int`
    """
    return get_terminal_size().columns - prefix_length - len(CONTINUOUS_LINE_POSTFIX)


def _add_linebreaks_between(lines):
    """
    Adds linebreaks between the given lines and returns a new list.
    
    Parameters
    ----------
    lines : `list<str>`
        Lines to add spaces between.
    
    Returns
    -------
    new_lines : `list<str>`
    """
    new_lines = []
    
    length = len(lines)
    if length:
        
        index = 0
        
        while True:
            line = lines[index]
            new_lines.append(line)
            
            index += 1
            if index >= length:
                break
            
            new_lines.append('\n')
            continue
    
    return new_lines


def iter_highlighted_buffer_parts(buffer, highlighter):
    """
    Iterates over the buffer parts and highlights them.
    
    This method is an iterable generator.
    
    Parameters
    ----------
    buffer : `list` of `str`
        Line buffer.
    highlighter : `None`, ``HighlightFormatterContext``
        The highlighter to use.
    
    Yields
    ------
    part : `str`
    """
    buffer = _add_linebreaks_between(buffer)
    
    if highlighter is None:
        yield from buffer
    
    else:
        yield from iter_highlight_code_lines(buffer, highlighter)


class DisplayState(RichAttributeErrorBaseType):
    """
    A current display state of ``EditorAdvanced``.
    
    Attributes
    ----------
    buffer : `list` of `str`
        Line buffer.
    content_width : `int`
        The screen's with.
    cursor_index : `int`
        The cursors index on the current line.
    cursor_line_index : `int`
        The cursor's line's index in the buffer.
    """
    __slots__ = ('buffer', 'content_width', 'cursor_index', 'cursor_line_index')
    
    def __new__(cls, buffer):
        """
        Creates a new display state.
        
        Parameters
        ----------
        buffer : `None`, `list` of `str`
            Line buffer.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        """
        buffer = _validate_buffer(buffer)
        if not buffer:
            buffer.append('')
        
        self = object.__new__(cls)
        self.buffer = buffer
        self.cursor_index = len(buffer[-1])
        self.cursor_line_index = len(buffer) - 1
        self.content_width = -1
        return self
    
    
    def __repr__(self):
        """
        Returns the representation of the display state.
        """
        repr_parts = ['<', type(self).__name__]
        
        repr_parts.append(' cursor_index = ')
        repr_parts.append(repr(self.cursor_index))
        
        repr_parts.append(', cursor_line_index = ')
        repr_parts.append(repr(self.cursor_line_index))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def copy(self):
        """
        Copies the display state.
        
        Returns
        -------
        new : ``DisplayState``
        """
        new = object.__new__(type(self))
        new.buffer = self.buffer.copy()
        new.cursor_index = self.cursor_index
        new.cursor_line_index = self.cursor_line_index
        new.content_width = -1
        return new
    
    
    def get_cursor_from_start_display_line_count(self, prefix_length):
        """
        Counts how much lines are from display position till the start of the display.
        
        Parameters
        ----------
        prefix_length : `int`
            Prefix length used for calculations when column width is changed.
        
        Returns
        -------
        count : `int`
        """
        content_width = self.content_width
        if (content_width == -1):
            # was never rendered
            return 0
        
        buffer = self.buffer
        cursor_index = self.cursor_index
        cursor_line_index = self.cursor_line_index
        
        count = _get_line_line_count(cursor_index, -1, content_width, prefix_length) - 1
        
        index = cursor_line_index - 1
        while index >= 0:
            count += _get_line_line_count(len(buffer[index]), -1, content_width, prefix_length)
            index -= 1
        
        return count
    
    
    def get_cursor_till_end_display_line_count(self, prefix_length, new_content_width):
        """
        Counts how much lines are from the display position till the end of the display.
        
        Parameters
        ----------
        prefix_length : `int`
            Prefix length used for calculations when column width is changed.
        new_content_width : `int`
            The new column width to use.
        
        Returns
        -------
        count : `int`
        """
        buffer = self.buffer
        cursor_index = self.cursor_index
        cursor_line_index = self.cursor_line_index
        content_width = self.content_width
        
        line = buffer[cursor_line_index]
        line_length = len(line)
        
        line_line_count = _get_line_line_count(line_length, content_width, new_content_width, prefix_length)
        line_cursor_index = _get_line_line_count(cursor_index, content_width, new_content_width, prefix_length)
        count = line_line_count - line_cursor_index
        
        index = cursor_line_index + 1
        limit = len(buffer)
        while index < limit:
            count += _get_line_line_count(len(buffer[index]), content_width, new_content_width, prefix_length)
            index += 1
        
        return count
    
    
    def get_start_till_end_display_line_count(self, prefix_length, new_content_width):
        """
        Counts how much lines are displayed.
        
        Parameters
        ----------
        prefix_length : `int`
            Prefix length used for calculations when column width is changed.
        new_content_width : `int`
            The new column width to use.
        
        Returns
        -------
        count : `int`
        """
        buffer = self.buffer
        count = 0
        content_width = self.content_width
        
        for line in buffer:
            count += _get_line_line_count(len(line), content_width, new_content_width, prefix_length)
        
        return count
    
    
    def jump_to_end(self, write_buffer, prefix_length, new_content_width):
        """
        Jumps to end of the displayed content.
        
        Parameters
        ----------
        write_buffer : `list<str>`
            Write buffer to extend.
        prefix_length : `int`
            Prefix length used to calculate line count.
        new_content_width : `int`
            The new column width to use.
        """
        if (self.content_width == -1):
            # self was never rendered
            return
        
        self._jump_to_end(write_buffer, prefix_length, new_content_width)
    
    
    def _jump_to_end(self, write_buffer, prefix_length, new_content_width):
        """
        Jumps to end of the displayed content.
        
        Called by ``.jump_to_end`` after checking pre-requirements.
        
        Parameters
        ----------
        write_buffer : `list<str>`
            Write buffer to extend.
        prefix_length : `int`
            Prefix length used to calculate line count.
        new_content_width : `int`
            The new column width to use.
        """
        _jump_vertical(write_buffer, 0, self.get_cursor_till_end_display_line_count(prefix_length, new_content_width))
    
    
    def clear(self, write_buffer, prefix_length, new_content_width):
        """
        Clears self from the editor.
        
        Parameters
        ----------
        write_buffer : `list<str>`
            Write buffer to extend.
        prefix_length : `int`
            Prefix length used to calculate line count.
        new_content_width : `int`
            The new column width to use.
        """
        if (self.content_width == -1):
            # self was never rendered
            return
        
        self._jump_to_end(write_buffer, prefix_length, new_content_width)
        self._clear_from_end(write_buffer, prefix_length, new_content_width)
    
    
    def _clear_from_end(self, write_buffer, prefix_length, new_content_width):
        """
        Clears self from the editor. Should be called after ``._jump_to_end``.
        
        Parameters
        ----------
        write_buffer : `list<str>`
            Write buffer to extend.
        prefix_length : `int`
            Prefix length used to calculate line count.
        new_content_width : `int`
            The new column width to use.
        """
        count = self.get_start_till_end_display_line_count(prefix_length, new_content_width)
        
        write_buffer.append(COMMAND_START_LINE)
        
        while True:
            write_buffer.append(COMMAND_CLEAR_LINE_WHOLE)
            
            count -= 1
            if count <= 0:
                break
            
            write_buffer.append(COMMAND_UP)
            continue
    
    
    def write_cursor(self, write_buffer, prefix_length):
        """
        Jumps the cursor to it's location.
        
        The output must be written out before.
        
        Parameters
        ----------
        write_buffer : `list<str>`
            Write buffer to extend.
        prefix_length : `int`
            Prefix length used to calculate line count.
        """
        # Jump back to cursor line
        write_buffer.append(
            create_command_up(self.get_cursor_till_end_display_line_count(prefix_length, self.content_width)),
        )
        write_buffer.append(COMMAND_START_LINE)
        write_buffer.append(
            create_command_right(_get_line_cursor_index(self.content_width, self.cursor_index, prefix_length)),
        )
    
    
    def write(self, write_buffer, prefix_length, prefix_initial, prefix_continuous, content_width):
        """
        Writes out the buffer.
        
        Parameters
        ----------
        write_buffer : `list<str>`
            Write buffer to extend.
        prefix_length : `int`
            Prefix length used to calculate line count.
        prefix_initial : `str`
            Prefix to use for the first line.
        prefix_continuous : `str`
            Prefix to use for additional lines.
        content_width : `int`
            Content width to use.
        """
        lines = self.get_line_render_intermediates(prefix_length, prefix_initial, prefix_continuous, content_width)
        
        line_count = len(lines)
        if not line_count:
            return
        
        write_buffer.append(COMMAND_FORMAT_RESET)
        index = 0
        
        while True:
            line = lines[index]
            index += 1
            
            write_buffer.append(COMMAND_START_LINE)
            
            for part in line.iter_parts():
                write_buffer.append(part)        
            
            if index == line_count:
                break
            
            write_buffer.append(COMMAND_DOWN)
            continue
    
    
    def write_difference(self, other, write_buffer, prefix_length, prefix_initial, prefix_continuous):
        """
        Writes out the difference between the two buffers..
        
        Parameters
        ----------
        other : `instance<type<self>>`
            Other instance to overwrite its buffer of.
        write_buffer : `list<str>`
            Write buffer to extend.
        prefix_length : `int`
            Prefix length used to calculate line count.
        prefix_initial : `str`
            Prefix to use for the first line.
        prefix_continuous : `str`
            Prefix to use for additional lines.
        """
        other_lines = other.get_line_render_intermediates(prefix_length, prefix_initial, prefix_continuous, -1)
        if not other_lines:
            return
        
        self_lines = self.get_line_render_intermediates(
            prefix_length, prefix_initial, prefix_continuous, other.content_width
        )
        
        current_line_index = other.get_cursor_from_start_display_line_count(prefix_length)
        last_written_line = None
        
        for line_index, self_line, other_line in zip_longest(
            range(max(len(other_lines), len(self_lines))),
            self_lines,
            other_lines,
        ):
            if other_line is None:
                _jump_vertical(write_buffer, current_line_index, line_index)
                current_line_index = line_index
                
                write_buffer.append(COMMAND_START_LINE)
                
                if (last_written_line is None):
                    write_buffer.append(COMMAND_FORMAT_RESET)
                
                for part in self_line.iter_parts():
                    write_buffer.append(part)
                
                last_written_line = self_line
                continue
            
            if self_line is None:
                _jump_vertical(write_buffer, current_line_index, line_index)
                current_line_index = line_index
                
                write_buffer.append(COMMAND_CLEAR_LINE_WHOLE)
                
                continue
            
            if self_line == other_line:
                continue
            
            # if we are different only write the difference.
            _jump_vertical(write_buffer, current_line_index, line_index)
            current_line_index = line_index
            
            write_buffer.append(COMMAND_START_LINE)
            
            last_command = None
            to_clear = 0
            jump_to = 0
            write_difference = False
            command_written = False
            
            item_index = 0
            self_items = self_line.parts
            other_items = other_line.parts
            
            # Skip same parts
            for item_index in range(item_index, min(len(self_items), len(other_items))):
                self_item = self_items[item_index]
                other_item = other_items[item_index]
                
                if self_item != other_item:
                    write_difference = True
                    break
                
                length = self_item[0]
                jump_to += length
                if not length:
                    last_command = self_item[1]
                
                continue
            
            # calculate additional shift in case we stopped on something.
            if not write_difference:
                self_length = 0
                self_part = ''
            
            else:
                self_length, self_part = self_item
                other_length, other_part = other_item
                
                # If we are not writing commands, cut the same part out
                if self_length and other_length:
                    cut_count, self_part, other_part = _cut_same_prefix(self_part, other_part)
                    jump_to += cut_count
                
                # Adjust clear count.
                to_clear += other_length - self_length
            
            # if we wanna jump, jump!!!
            if jump_to > 0:
                write_buffer.append(create_command_right(jump_to))
            
            # if wanna write, write!!
            if self_part:
                if self_length:
                    if (last_command is not None):
                        write_buffer.append(last_command)
                    elif (last_written_line is None):
                        write_buffer.append(COMMAND_FORMAT_RESET)
                        
                
                write_buffer.append(self_part)
                command_written = True
            
            # Increment index by 1 to skip the current element next time.
            item_index += 1
            
            # write self parts
            if len(self_items) > item_index:
                if not command_written:
                    write_buffer.append(COMMAND_FORMAT_RESET)
                
                for self_length, self_part in islice(self_items, item_index, None):
                    to_clear -= self_length
                    write_buffer.append(self_part)
            
            # calculate how much to clear
            for other_length, other_part in islice(other_items, item_index, None):
                to_clear += other_length
            
            # fill up space if required.
            if to_clear > 0:
                write_buffer.append(COMMAND_CLEAR_LINE_FROM_CURSOR)
            
            last_written_line = self_line
            continue
        
        # write cursor
        _jump_vertical(
            write_buffer,
            current_line_index,
            self.get_cursor_from_start_display_line_count(prefix_length),
        )
        
        if last_written_line is None:
            cursor_index = _get_line_cursor_index(other.content_width, other.cursor_index, prefix_length)
        else:
            cursor_index = last_written_line.get_length()
        
        _jump_horizontal(
            write_buffer,
            cursor_index,
            _get_line_cursor_index(other.content_width, self.cursor_index, prefix_length),
        )
    
    
    def get_line_render_intermediates(self, prefix_length, prefix_initial, prefix_continuous, content_width):
        """
        Gets line rendering intermediates that contain the content to write but are not written yet.
        
        Parameters
        ----------
        prefix_length : `int`
            Prefix length used to calculate line count.
        prefix_initial : `str`
            Prefix to use for the first line.
        prefix_continuous : `str`
            Prefix to use for additional lines.
        content_width : `int`
            Content width to use.
        
        Returns
        -------
        lines : `list<LineRenderIntermediate>`
        """
        if content_width == -1:
            content_width = self.content_width
        else:
            if content_width < 1:
                return []
        
            self.content_width = content_width
        
        lines = []
        
        line_index = 0
        buffer = self.buffer
        line_count = len(buffer)
        leftover_line_length = content_width
        last_command = COMMAND_FORMAT_RESET
        
        line = LineRenderIntermediate(prefix_length, prefix_initial)
        lines.append(line)
        
        for part in iter_highlighted_buffer_parts(buffer, DEFAULT_ANSI_HIGHLIGHTER):
            # if the part ends with `\n`, remove it and remember it for later
            if part.endswith('\n'):
                is_line_break = True
                part = part[:-2]
            else:
                is_line_break = False
            
            # write line if any. If the part is a command, it will have 0 visible length, so ignore it
            # else only write as much as we can, then break line and continue. 
            if part:
                if is_command(part):
                    line.add_command(part)
                    last_command = part
                
                else:
                    while True:
                        leftover_line_length -= len(part)
                        if leftover_line_length >= 0:
                            line.add_part(part)
                            break
                        
                        line.add_part(part[:leftover_line_length])
                        line.add_part(CONTINUOUS_LINE_POSTFIX)
                        part = part[leftover_line_length:]
                        
                        # Create new line
                        line = LineRenderIntermediate(prefix_length, EMPTY_LINE_PREFIX_CHARACTER * prefix_length)
                        lines.append(line)
                        leftover_line_length = content_width
                        line.add_command(last_command)
                        continue
            
            # If we wrote a line break:
            # - quit if last
            # - jump to next line and write prefix if we have more
            if is_line_break:
                # At last line skip line break
                line_index += 1
                if line_index == line_count:
                    break
                
                
                line = LineRenderIntermediate(prefix_length, prefix_continuous)
                lines.append(line)
                leftover_line_length = content_width
                continue
            
            continue
    
        return lines



__all__ = ()

from ....utils import RichAttributeErrorBaseType


class LineRenderIntermediate(RichAttributeErrorBaseType):
    """
    Represents intermediate information about how a line is rendered.
    
    Attributes
    ----------
    parts : `list<(int, str)>`
        The parts the line consists of.
    """
    __slots__ = ('parts',)
    
    def __new__(cls, prefix_length, prefix):
        """
        Creates a new line render instance.
        
        Parameters
        ----------
        prefix_length : `int`
            The length of the prefix.
        prefix : `str`
            The prefix.
            Prefixes might contain commands in them, so their length and value is separated.
        """
        self = object.__new__(cls)
        self.parts = [(prefix_length, prefix)]
        return self
    
    
    def add_command(self, command):
        """
        Adds a command to the line render.
        The visual representation of these lines equal to `0`.
        
        Parameters
        ----------
        command : `str`
            The command to add.
        """
        self.parts.append((0, command))
    
    
    def add_part(self, part):
        """
        Adds a part to the line render.
        The visual representation of these lines equal to their length.
        
        Parameters
        ----------
        part : `str`
            The part to add.
        """
        self.parts.append((len(part), part))
    
    
    def __eq__(self, other):
        """Returns whether the two renders are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self.parts == other.parts
    
    
    def __repr__(self):
        """Returns the line render's representation."""
        repr_parts = ['<', type(self).__name__]
        
        repr_parts.append(' parts = ')
        repr_parts.append(repr(self.parts))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def iter_parts(self):
        """
        Iterates over the added parts. This also includes commands too.
        
        This method is an iterable generator.
        
        Yields
        ------
        part : `str`
        """
        for element in self.parts:
            yield element[1]
    
    
    def iter_items(self):
        """
        Iterates over the added items. This includes whether a part is a command and the part itself.
        
        This method is an iterable generator.
        
        Yields
        ------
        length, part : (int, str)
        """
        yield from self.parts
    
    
    def get_length(self):
        """
        Returns the length of the line. This does not equal to its part count.
        
        Returns
        -------
        length : `int`
        """
        return sum(part[0] for part in self.parts)



__all__ = ('FormData', )

from io import IOBase
from urllib.parse import urlencode

from ..utils import IgnoreCaseMultiValueDictionary, MultiValueDictionary, RichAttributeErrorBaseType, to_json

from .headers import CONTENT_LENGTH, CONTENT_TRANSFER_ENCODING, CONTENT_TYPE
from .multipart import BytesPayload, MultipartWriter, create_payload


FORM_DATA_FIELD_TYPE_NONE = 0
FORM_DATA_FIELD_TYPE_JSON = 1


class FormDataField(RichAttributeErrorBaseType):
    """
    Attributes
    ----------
    headers : `IgnoreCaseMultiValueDictionary<str, str>`
        The field specific headers.
    type : `int`
        The field's type.
    type_options : `MultiValueDictionary<str, str>`
        Additional information used by the created ``PayloadBase`` when the field is generated.
    value : object
        The field's value.
    """
    __slots__ = ('headers', 'type', 'type_options', 'value')
    
    def __new__(cls, field_type, type_options, headers, value):
        """
        Creates a new form data field.
        
        Parameters
        ----------
        field_type : `int`
            The field's type.
        type_options : `MultiValueDictionary<str, str>`
            Additional information used by the created ``PayloadBase`` when the field is generated.
        headers : `IgnoreCaseMultiValueDictionary<str, str>`
            The field specific headers.
        value : `object`
            The field's value.
        """
        self = object.__new__(cls)
        self.type = field_type
        self.type_options = type_options
        self.headers = headers
        self.value = value
        return self
    
    
    def get_value(self):
        """
        Returns the value of the form data field after applying type specific conversion.
        
        Returns
        -------
        value : `object`
        """
        field_type = self.type
        value = self.value
        
        # none
        if field_type == FORM_DATA_FIELD_TYPE_NONE:
            return value
        
        # json
        if field_type == FORM_DATA_FIELD_TYPE_JSON:
            return to_json(value)
        
        # unknown
        return value
    
    
    def __repr__(self):
        """Returns the field's representation."""
        repr_parts = ['<', type(self).__name__]
        
        # type
        field_type = self.type
        if field_type == FORM_DATA_FIELD_TYPE_NONE:
            type_name = 'none'
        elif field_type == FORM_DATA_FIELD_TYPE_JSON:
            type_name = 'json'
        else:
            type_name = 'unknown'
        
        repr_parts.append(' type = ')
        repr_parts.append(type_name)
        
        # type_options
        type_options = self.type_options
        if type_options:
            repr_parts.append(', type_options = ')
            repr_parts.append(repr(type_options))
        
        # headers
        headers = self.headers
        if headers:
            repr_parts.append(', headers = ')
            repr_parts.append(repr(headers))
        
        # value
        repr_parts.append(', value = ')
        value = self.value
        if isinstance(value, (bytes, bytearray, memoryview)) and len(value) > 10000:
            repr_parts.append('binary<length = ')
            repr_parts.append(str(len(value)))
            repr_parts.append('>')
        
        else:
            repr_parts.append(repr(value))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two fields are equal."""
        if type(self) is not type(other):
            return False
        
        # type
        if self.type != other.type:
            return False
        
        # type_options
        if self.type_options != other.type_options:
            return False
        
        # headers
        if self.headers != other.headers:
            return False
        
        # value
        if self.value != other.value:
            return False
        
        return True


class FormData(RichAttributeErrorBaseType):
    """
    Helper class for `multipart/form-data` and `application/x-www-form-urlencoded` body generation.
    
    Attributes
    ----------
    fields : `list<FormDataField>`
        The fields of the form data.
    
    multipart : `bool`
        Whether the form data is `multipart/form-data` and not `application/x-www-form-urlencoded` type.
    
    quote_fields : `bool`
        Whether type option field values should be quoted.
    """
    __slots__ = ('fields', 'multipart', 'quote_fields')
    
    def __new__(cls, quote_fields = True):
        """
        Creates a new a form data.
        
        Parameters
        ----------
        quote_fields : `bool` = `True`, Optional
            Whether type option field values should be quoted.
        """
        self = object.__new__(cls)
        self.fields = []
        self.multipart = False
        self.quote_fields = quote_fields
        return self
    
    
    @classmethod
    def from_fields(cls, fields):
        """
        Creates a new form data from the given fields.
        
        Parameters
        ----------
        fields : `dict<str, object> | list<(str, object)> | IOBase`
            The fields to convert to form data.
        
        Returns
        -------
        self : `instance<cls>`
            The created form data instance.
        
        Raises
        ------
        TypeError
            Received unhandleable field.
        """
        if isinstance(fields, dict):
            fields = list(fields.items())
        
        elif isinstance(fields, list) or isinstance(fields, tuple):
            fields = [*fields]
        
        else:
            fields = [fields]
        
        self = cls()
        
        while fields:
            field = fields.pop(0)
            fiend_type = type(field)
            
            if issubclass(fiend_type, IOBase):
                self.add_field(getattr(field, 'name', 'unknown'), field)
            
            elif issubclass(fiend_type, tuple) and len(field) == 2:
                self.add_field(*field)
            
            else:
                raise TypeError(
                    f'`{type(self).__name__}.from_fields` got unhandleable field: {fiend_type.__name__}; {field!r}.'
                )
        
        return self
    
    
    def add_json(self, name, value):
        """
        Adds a json field to the form data.
        
        Parameters
        ----------
        name : `str`
            The field's name.
        value : `object`
            The field's value.
        """
        self.multipart = True
        
        type_options = MultiValueDictionary()
        type_options['name'] = name
        
        headers = IgnoreCaseMultiValueDictionary()
        headers[CONTENT_TYPE] = 'application/json'
        
        self.fields.append(FormDataField(FORM_DATA_FIELD_TYPE_JSON, type_options, headers, value))
    
    
    def add_field(self, name, value, content_type = None, file_name = None, transfer_encoding = None):
        """
        Adds a field to the form data.
        
        Parameters
        ----------
        name : `str`
            The field's name.
        value : `object`
            The field's value.
        content_type : `None`, `str` = `None`, Optional
            The field's content type.
        file_name : `None`, `str` = `None`, Optional
            The field's name. If not given or given as `None` (so by the default), and the given `value` is an `IOBase`
            instance then tries to use that's name.
        transfer_encoding : `None`, `str` = `None`, Optional
            The field's transfer encoding.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        """
        if isinstance(value, IOBase):
            self.multipart = True
        
        elif isinstance(value, (bytes, bytearray, memoryview)):
            if (file_name is None) and (transfer_encoding is None):
                file_name = name
        
        type_options = MultiValueDictionary()
        type_options['name'] = name
        
        if (file_name is None) and isinstance(value, IOBase):
            file_name = getattr(value, 'name', name)
        
        if (file_name is not None):
            if not isinstance(file_name, str):
                raise TypeError(
                    f'`file_name` can be `None`, `str`, got {type(file_name).__name__}; {file_name!r}.'
                )
            
            type_options['file_name'] = file_name
            self.multipart = True
        
        headers = IgnoreCaseMultiValueDictionary()
        if (content_type is not None):
            if not isinstance(content_type, str):
                raise TypeError(
                    f'`content_type` can be `None`, `str`, got {type(content_type).__name__}; {content_type!r}.'
                )
            
            headers[CONTENT_TYPE] = content_type
            self.multipart = True
        
        if (transfer_encoding is not None):
            if not isinstance(transfer_encoding, str):
                raise TypeError(
                    f'`transfer_encoding` can be `None`, `str`, got: {type(transfer_encoding).__name__}; '
                    f'{transfer_encoding!r}.'
                )
            
            headers[CONTENT_TRANSFER_ENCODING] = transfer_encoding
            self.multipart = True
        
        self.fields.append(FormDataField(FORM_DATA_FIELD_TYPE_NONE, type_options, headers, value))
    
    
    def _generate_form_data(self, encoding):
        """
        Generates `multipart/form-data` payload from the form data fields.
        
        Parameters
        ----------
        encoding : `str`
            The encoding to use to encode the form data's fields.
        
        Returns
        -------
        payload : ``MultipartWriter``
            The generated payload.
        
        Raises
        ------
        TypeError
            Cannot serialize a field.
        RuntimeError
            - If a field's content has unknown content-encoding.
            - If a field's content has unknown content-transfer-encoding.
        """
        writer = MultipartWriter('form-data')
        for field in self.fields:    
            try:
                type_options = field.type_options
                headers = field.headers
                value = field.get_value()
            
                payload_keyword_parameters = {
                    'headers': headers,
                    'encoding': encoding,
                }
                
                try:
                    content_type = headers[CONTENT_TYPE]
                except KeyError:
                    pass
                else:
                    payload_keyword_parameters['content_type'] = content_type
                    
                    if type_options:
                        payload_keyword_parameters.update(type_options.kwargs())
                
                part = create_payload(value, payload_keyword_parameters)
            
            except Exception as err:
                raise TypeError(
                    f'Can not serialize field: {field!r}.'
                ) from err
            
            if type_options:
                part.set_content_disposition_header('form-data', type_options.kwargs(), self.quote_fields)
                part.headers.pop_all(CONTENT_LENGTH, None)
            
            writer.append_payload(part)
        
        return writer
    
    
    def _generate_form_urlencoded(self, encoding):
        """
        Generates `application/x-www-form-urlencoded` payload from the form data's fields.
        
        Parameters
        ----------
        encoding : `str`
            The encoding to use to encode the form data's fields.
        
        Returns
        -------
        payload : ``BytesPayload``
            The generated payload.
        """
        data = []
        for field in self.fields:
            data.append((field.type_options['name'], field.get_value()))
        
        if encoding == 'utf-8':
            content_type = 'application/x-www-form-urlencoded'
        else:
            content_type = f'application/x-www-form-urlencoded; charset={encoding}'
        
        return BytesPayload(urlencode(data, doseq = True, encoding = encoding).encode(), {'content_type': content_type})
    
    
    def generate_form(self, encoding = 'utf-8'):
        """
        Gets the payload of the form data..
        
        Parameters
        ----------
        encoding : `str` = `'utf-8'`, Optional
            The encoding to use to encode the form data's fields.
        
        Returns
        -------
        payload : ``BytesPayload``, ``MultipartWriter``
            The generated payload.
        
        Raises
        ------
        TypeError
            Cannot serialize a field.
        """
        if self.multipart:
            return self._generate_form_data(encoding)
        else:
            return self._generate_form_urlencoded(encoding)
    
    
    def __repr__(self):
        """Returns the representation of the form data."""
        repr_parts = ['<', type(self).__name__]
        
        # fields
        fields = self.fields
        if not fields:
            field_added = False
        else:
            repr_parts.append(' fields = ')
            repr_parts.append(repr(fields))
            
            field_added = True
        
        # multipart
        multipart = self.multipart
        if multipart:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' multipart = ')
            repr_parts.append(repr(multipart))
        
        # quote_fields
        if field_added:
            repr_parts.append(',')
        else:
            field_added = True
        
        repr_parts.append(' quote_fields = ')
        repr_parts.append(repr(self.quote_fields))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two form datas are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.fields != other.fields:
            return False
        
        if self.multipart != other.multipart:
            return False
        
        if self.quote_fields != other.quote_fields:
            return False
        
        return True
    
    
    def __bool__(self):
        """Returns whether the form data has any fields."""
        return True if self.fields else False



__all__ = ()

from ..rich_attribute_error import RichAttributeErrorBaseType

from .frame_ignoring import should_keep_frame, should_keep_frame_from_filter
from .frame_proxy import FrameProxyVirtual
from .repeat_strategies import get_repeat_with_strategy_bot


FRAME_GROUP_TYPE_NONE = 0
FRAME_GROUP_TYPE_SINGLES = 1
FRAME_GROUP_TYPE_REPEAT = 2


class FrameGroup(RichAttributeErrorBaseType):
    """
    Represents grouped frames.
    
    Attributes
    ----------
    frames : `None | list<FrameProxyBase>`
        The grouped up frames.
    repeat_count : `int`
        How much times the wrapped frames are grouped.
    type : `int`
        Information about how the frames are grouped.
    """
    __slots__ = ('frames', 'repeat_count', 'type')
    
    def __new__(cls):
        """
        Creates a new frame group.
        """
        self = object.__new__(cls)
        self.frames = None
        self.type = FRAME_GROUP_TYPE_NONE
        self.repeat_count = 0
        return self
    
    
    @classmethod
    def _create_repeated(cls, frames, repeat_count):
        """
        Creates a repeated frame group.
        
        Parameters
        ----------
        frames : `list<FrameProxyBase>`
            The stored frames by the group.
        repeat_count : `int`
            How much times is the group repeated.
        
        Returns
        -------
        self : ``FrameGroup``
        """
        self = object.__new__(cls)
        self.frames = frames
        self.repeat_count = repeat_count
        self.type = FRAME_GROUP_TYPE_REPEAT
        return self
    
    
    def __eq__(self, other):
        """Returns whether the two frame groups are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal(other)
    
    
    def _is_equal(self, other):
        """
        Returns whether the two frame groups are equal.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other instance.
        
        Returns
        -------
        is_equal : `bool`
        """
        if len(self) != len(other):
            return False
        
        for self_frame, other_frame in zip(self.iter_frames(), other.iter_frames()):
            if self_frame != other_frame:
                return False
        
        return True
    
    
    def __mod__(self, other):
        """Returns whether the two frame groups are alike."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_alike(other)
    
    
    def __rmod__(self, other):
        """Returns whether the two frame groups are alike."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_alike(other)
    
    
    def _is_alike(self, other):
        """
        Returns whether the frame groups are alike.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other instance.
        
        Returns
        -------
        is_alike : `bool`
        """
        if len(self) != len(other):
            return False
        
        for self_frame, other_frame in zip(self.iter_frames(), other.iter_frames()):
            if not (self_frame % other_frame):
                return False
        
        return True
        
    
    def __len__(self):
        """Returns how much frames are in the group."""
        frames = self.frames
        if frames is None:
            return 0
        
        return len(frames) * self.repeat_count
    
    
    def __repr__(self):
        """Returns the frame group's representation."""
        repr_parts = ['<', type(self).__name__]
        
        field_added = False
        
        repeat_count = self.repeat_count
        if repeat_count > 1:
            field_added = True
            repr_parts.append(' repeat_count = ')
            repr_parts.append(repr(repeat_count))
        
        frames = self.frames
        if (frames is not None):
            if field_added:
                repr_parts.append(',')
            
            repr_parts.append(' frames = ')
            repr_parts.append(repr(frames))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __bool__(self):
        """Returns whether the frame group has any elements."""
        return self.type != FRAME_GROUP_TYPE_NONE
    
    
    def iter_frames(self):
        """
        Iterates over the frames of the frame group.
        
        This method is an iterable generator.
        
        Yields
        ------
        frame : ``FrameProxyBase``
        """
        frames = self.frames
        if (frames is not None):
            for index in range(self.repeat_count):
                yield from frames
    
    
    def iter_exhaust_frames(self):
        """
        Iterates over the frames of the frame group, exhausting its own.
        
        This method is an iterable generator.
        
        Yields
        ------
        frame : ``FrameProxyBase``
        """
        try:
            yield from self.iter_frames()
        finally:
            self.clear()
    
    
    def iter_frames_no_repeat(self):
        """
        Iterates over the frames of the frame group without repeating repeated blocks.
        It may still happen that non-unique frames are yielded.
        
        This method is an iterable generator.
        
        Yields
        ------
        frame : ``FrameProxyBase``
        """
        frames = self.frames
        if (frames is not None):
            yield from frames
    
    
    def clear(self):
        """
        Clears the frame group.
        """
        self.frames = None
        self.repeat_count = 0
        self.type = FRAME_GROUP_TYPE_NONE
    
    
    def copy(self):
        """
        Copies the frame group.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        frames = self.frames
        if (frames is not None):
            frames = frames.copy()
        new.frames = frames
        new.repeat_count = self.repeat_count
        new.type = self.type
        return new
    
    
    def drop_variables(self):
        """
        Drops the variables of the frame group.
        """
        frames = self.frames
        if (frames is not None):
            self.frames = [FrameProxyVirtual(frame, with_variables = False) for frame in frames]
    
    
    def copy_without_variables(self):
        """
        Copies the frame group without variables.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        frames = self.frames
        if (frames is not None):
            frames = [FrameProxyVirtual(frame, with_variables = False) for frame in frames]
        new.frames = frames
        new.repeat_count = self.repeat_count
        new.type = self.type
        return new
    
    
    def has_variables(self):
        """
        Returns whether any frames inherited variables on creation.
        
        Returns
        -------
        has_variables : `bool`
        """
        frames = self.frames
        if (frames is not None):
            for frame in frames:
                if frame.has_variables():
                    return True
        
        return False
    
    
    def drop_ignored_frames(self, *, filter = None):
        """
        Drops the frames that should be ignored.
        
        Parameters
        ----------
        filter : `None | callable` = `None`, Optional (Keyword only)
            Additional filter to check whether a frame should be shown.
        """
        frames = self.frames
        if (frames is None):
            return
        
        frames = [frame for frame in frames if should_keep_frame(frame, filter = filter)]
        if frames:
            self.frames = frames
            return
        
        self.type = FRAME_GROUP_TYPE_NONE
        self.repeat_count = 0
        self.frames = None
    
    
    def apply_frame_filter(self, filter):
        """
        Keeps the frames that pass the given `filter`.
        
        Parameters
        ----------
        filter : `callable`
            Filter to check whether a frame should be kept.
        """
        frames = self.frames
        if (frames is None):
            return
        
        frames = [frame for frame in frames if should_keep_frame_from_filter(frame, filter)]
        if frames:
            self.frames = frames
            return
        
        self.type = FRAME_GROUP_TYPE_NONE
        self.repeat_count = 0
        self.frames = None
    
    
    def try_add_frame(self, frame):
        """
        Tries to add the given frame to self. On success returns `True`.
        
        Parameters
        ----------
        frame : ``FrameProxyBase``
            The frame to add.
        
        Returns
        -------
        success : `bool`
        """
        group_type = self.type
        
        if group_type == FRAME_GROUP_TYPE_NONE:
            self.frames = [frame]
            self.repeat_count = 1
            
            if frame.alike_count > 1:
                group_type = FRAME_GROUP_TYPE_REPEAT
            else:
                group_type = FRAME_GROUP_TYPE_SINGLES
            self.type = group_type
            return True
        
        if group_type == FRAME_GROUP_TYPE_SINGLES:
            if frame.alike_count > 1:
                return False
            
            self.frames.append(frame)
            return True
        
        
        if group_type == FRAME_GROUP_TYPE_REPEAT:
            if frame.alike_count <= 1:
                return False
            
            self.frames.append(frame)
            return True
        
        # No more cases
        return False
    
    
    def try_merge(self, other):
        """
        Tries to merge self with an other frame group.
        
        Parameters
        ----------
        other : `instance<cls>`
            The frame group to merge self with.
        
        Returns
        -------
        merged : `bool`
        """
        if (self.repeat_count != 1) or (other.repeat_count != 1):
            return False
        
        self_group_type = self.type
        other_group_type = other.type
        
        # single + single
        if self_group_type == FRAME_GROUP_TYPE_SINGLES and other_group_type == FRAME_GROUP_TYPE_SINGLES:
            should_merge = True
        
        # single + repeat -> short
        elif self_group_type == FRAME_GROUP_TYPE_SINGLES and other_group_type == FRAME_GROUP_TYPE_REPEAT:
            if len(other.frames) <= 2:
                should_merge = True
            else:
                should_merge = False
        
        # repeat + single -> short
        elif self_group_type == FRAME_GROUP_TYPE_REPEAT and other_group_type == FRAME_GROUP_TYPE_SINGLES:
            if len(self.frames) <= 2:
                should_merge = True
            else:
                should_merge = False
        
        # `repeat + repeat` cannot happen
        else:
            should_merge = False
        
        if should_merge:
            self.frames.extend(other.iter_exhaust_frames())
            self.type = FRAME_GROUP_TYPE_SINGLES
        
        return should_merge
    
    
    def iter_separate_repeated(self):
        """
        Iterates over self producing repeated groups.
        
        Yields
        ------
        checked : `bool`
        frame_group : `instance<type<self>>`
        """
        # Yield self if non-repeat or if already processed repeat.
        if self.type != FRAME_GROUP_TYPE_REPEAT or self.repeat_count > 1:
            yield True, self
            return
        
        # Find repeat
        repeat_range = get_repeat_with_strategy_bot(self.frames)
        if repeat_range is None:
            self.type = FRAME_GROUP_TYPE_SINGLES
            yield True, self
            return
        
        start_shift, chunk_size, repeat = repeat_range
        frames = self.frames
        
        # fore part
        if start_shift > 0:
            yield False, self._create_repeated(frames[0 : start_shift], 1)
        
        # mid part
        yield True, self._create_repeated(frames[start_shift : start_shift + chunk_size], repeat)
        
        # last part
        frames_length = len(frames)
        end_shift = start_shift + chunk_size * repeat
        if frames_length > end_shift:
            yield False, self._create_repeated(frames[end_shift : frames_length], 1)
    
    
    def get_last_frame(self):
        """
        Returns the last frame of the frame group.
        
        Returns
        -------
        frame : `None | FrameProxyBase`
        """
        frames = self.frames
        if frames is not None:
            return frames[-1]



__all__ = ('ConnectorTCP',)

from functools import partial as partial_func
from http import HTTPStatus
from socket import AddressFamily, SOCK_STREAM as SOCKET_TYPE_STREAM
from ssl import (
    CertificateError as SSLCertificateError, SSLContext, SSLError, create_default_context as create_default_ssl_context
)
from warnings import warn

from ..core import CancelledError, Future, SSLBidirectionalTransportLayer, Task
from ..utils import CauseGroup, IgnoreCaseMultiValueDictionary
from ..web_common.exceptions import ProxyError
from ..web_common.headers import METHOD_CONNECT
from ..web_common.helpers import set_tcp_nodelay
from ..web_common.http_protocol import HttpReadWriteProtocol

from .client_request import ClientRequest
from .connection import Connection
from .connector_base import ConnectorBase
from .constants import SSL_CONTEXT_UNVERIFIED
from .host_info import HostInfo
from .host_info_basket import HostInfoBasket
from .ssl_fingerprint import SSLFingerprint


class ConnectorTCP(ConnectorBase):
    """
    Tcp connector.
    
    Attributes
    ----------
    cookies : `http.cookies.SimpleCookie`
        Cookies of the connection.
    
    clean_up_handle : `None | TimerWeakHandle`
        Weak handle, which cleans up the timed out connections of the connector.
    
    closed : `bool`
        Whether the connector is closed.
    
    family : `AddressFamily`
        Address family of the created socket if any.
    
    force_close : `bool`
        Whether after each request (and between redirects) the connections should be closed.
    
    host_info_basket_cache : `dict<(None | str, None | int), HostInfoBasket)`
        Cached resolved host information.
    
    local_address : `None | (None | str, None | int))`
        Can be given as a `tuple` (`local_host`, `local_port`) to bind created sockets locally.
    
    loop : ``EventThread``
        The event loop to what the connector is bound to.
    
    protocols_by_host : `dict<ConnectionKey, ConnectionBasket>`
        Protocols for each host.
    
    resolve_host_tasks_and_waiters : `dict<(None | str, None | int), \
            (Task<.resolve>, list<Future<(None | HostInfoBasket, None | BaseException)>>)>`
        Active host info resolving tasks and waiters.
    
    ssl_context : `None | SSLContext`
        The connection's ssl type.
    
    ssl_fingerprint : `None | SSLFingerprint`
        Alternative way to accept ssl or to block it depending whether the fingerprint is the same or changed.
    
    Notes
    -----
    Connectors support weakreferencing.
    """
    __slots__ = (
        'family', 'host_info_basket_cache', 'local_address', 'resolve_host_tasks_and_waiters', 'ssl_context',
        'ssl_fingerprint'
    )
    
    def __new__(
        cls,
        loop,
        *deprecated,
        family = AddressFamily.AF_UNSPEC,
        force_close = False,
        local_address = None,
        ssl = ...,
        ssl_context = None,
        ssl_fingerprint = None,
    ):
        """
        Creates a new tcp connector with the given parameters.
        
        Parameters
        ----------
        loop : ``EventThread``
            The event loop to what the connector is bound to.
        
        family : `AddressFamily` = `AddressFamily.AF_UNSPEC`, Optional (keyword only)
            Address family of the created socket if any
        
        force_close : `bool` = `False`, Optional (keyword only)
            Whether after each request (and between redirects) the connections should be closed.
        
        local_address : `None | (None | str, None | int)` = `None`, Optional (keyword only)
            Can be given as a `tuple` (`local_host`, `local_port`) to bind created sockets locally.
        
        ssl_context : `None | SSLContext`, Optional (Keyword only)
            SSL context to be used by the connector.
        
        ssl_fingerprint : `None | SSLFingerprint`, Optional (Keyword only)
            SSL finger print to use by the connector.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        """
        # deprecated
        deprecated_length = len(deprecated)
        if deprecated_length:
            warn(
                (
                    f'The `force_close`, `proxy_headers`, `local_address` and `ssl` parameters in '
                    f'`{cls.__name__}.__new__` are moved to be keyword only. '
                    f'Support for positional is deprecated and will be removed in 2025 August.'
                ),
                FutureWarning,
                stacklevel = 2,
            )
            
            family = deprecated[0]
            
            if deprecated_length > 1:
                ssl = deprecated[1]
            
            if deprecated_length > 2:
                local_address = deprecated[2]
            
            if deprecated_length > 3:
                force_close = deprecated[3]
        
        
        if (ssl is not ...):
            warn(
                (
                    f'`{cls.__name__}.__new__`\'s `ssl` parameter is deprecated '
                    f'and scheduled for removal in 2025 August. '
                    f'Please use either the `ssl_context` or the `ssl_fingerprint` parameters depending on your needs.'
                ),
                FutureWarning,
                stacklevel = 3,
            )
            
            if ssl is None:
                ssl_context = None
                ssl_fingerprint = None
            
            elif isinstance(ssl, SSLContext):
                ssl_context = ssl
                ssl_fingerprint = None
            
            elif isinstance(ssl, SSLFingerprint):
                ssl_context = None
                ssl_fingerprint = ssl
            
            elif isinstance(ssl, bool):
                ssl_context = create_default_ssl_context()
                ssl_fingerprint = None
                
            else:
                raise TypeError(
                    f'`ssl` can be `None`, `bool`, `{SSLContext.__name__}`, `{SSLFingerprint.__name__}`. '
                    f'Got {type(ssl).__name__}; {ssl!r}.'
                )
        
        
        self = ConnectorBase.__new__(cls, loop, force_close = force_close)
        
        self.family = family
        self.host_info_basket_cache = {}
        self.local_address = local_address
        self.resolve_host_tasks_and_waiters = {}
        self.ssl_context = ssl_context
        self.ssl_fingerprint = ssl_fingerprint
        
        return self
    
    
    def close(self):
        """
        Closes the connector, it's dns lookup events and it's connections.
        """
        for task, waiters in self.resolve_host_tasks_and_waiters.values():
            task.cancel()
        
        ConnectorBase.close(self)
    
    
    async def resolve_host(self, host, port, waiters):
        """
        Resolves a host and returns it's result.
        
        This method is a coroutine.
        
        Parameters
        ----------
        host : `None | str`
            To what network interfaces should the connection be bound.
        port : `None | int`
            The port of the `host`.
        
        Returns
        -------
        result : `(None | HostInfoBasket, None | BaseException)`
            A host info container containing the resolved addresses. And an occurred exception if any.
        """
        result = (None, None)
        
        try:
            try:
                infos = await self.loop.get_address_info(host, port, type = SOCKET_TYPE_STREAM, family = self.family)
            except GeneratorExit:
                raise
            
            except CancelledError:
                raise
            
            except BaseException as exception:
                result = (None, exception)
            
            else:
                new_host_info_basket = HostInfoBasket.from_address_infos(host, infos)
                self.host_info_basket_cache[host, port] = new_host_info_basket
                result = (new_host_info_basket, None)
        
        finally:
            try:
                del self.resolve_host_tasks_and_waiters[host, port]
            except KeyError:
                pass
            
            for waiter in waiters:
                waiter.set_result_if_pending(result)
    
    
    def get_resolve_host_waiter(self, host, port):
        """
        Resolves addresses for the given `host` and `port`.
        Returns a future that can be awaited to retrieve the result.
        
        Parameters
        ----------
        host : `None | str`
            To what network interfaces should the connection be bound.
        port : `None | int`
            The port of the `host`.
        
        Returns
        -------
        waiter : `Future<(HostInfoBasket, BaseException)>`
            A waiter future that can be awaited to retrieve the resolve's result.
        """
        try:
            task, waiters = self.resolve_host_tasks_and_waiters[host, port]
        except KeyError:
            waiters = []
            task = Task(self.loop, self.resolve_host(host, port, waiters))
            self.resolve_host_tasks_and_waiters[host, port] = (task, waiters)
        
        waiter = Future(self.loop)
        waiters.append(waiter)
        
        return waiter
    
    
    async def resolve_host_iterator(self, request):
        """
        Yields host infos to which the request can connect to.
        
        This method is a coroutine generator.
        
        Parameters
        ----------
        request : ``ClientRequest``
            Respective request, which requires a connection.
        
        Yields
        ------
        host_info : ``HostInfo``
            Resolved host information.
        
        Raises
        ------
        ConnectionError
            No hosts were resolved.
        BaseException
            Any other cached exception.
        """
        url = request.url
        host = url.raw_host
        port = request.port
        
        # If host is an ip return
        if url.is_host_ip():
            yield HostInfo.from_ip(host, port, self.family)
            return
        
        # Strip extra dots from the end, keep one.
        # This is actually done now in url parsing because double dot is invalid.
        
        # Get old host info basket
        try:
            old_host_info_basket = self.host_info_basket_cache[host, port]
        except KeyError:
            old_host_info_basket = None
            expired = True
        else:
            expired = old_host_info_basket.is_expired()
        
        # if old host info basket expired create new resolve task.
        if expired:
            waiter = self.get_resolve_host_waiter(host, port)
        
        if (old_host_info_basket is not None):
            for host_info in old_host_info_basket.iter_next_rotation():
                yield host_info
        
        if not expired:
            return
        
        new_host_info_basket, exception = await waiter
        
        # Yield host infos that we did not yield before.
        if (new_host_info_basket is not None):
            for host_info in new_host_info_basket.iter_next_rotation():
                if (old_host_info_basket is None) or (host_info not in old_host_info_basket):
                    yield host_info
            
            return
        
        # Raise any occurred exception.
        if (exception is not None):
            if not isinstance(exception, OSError):
                raise exception
            
            raise ConnectionError(request.connection_key, 'Resolve task failed.') from exception
            
        # If neither host info basket & exception is returned then the resolve task was cancelled.
        raise ConnectionError(request.connection_key, 'Resolve task cancelled.')
    
    
    async def create_connection(self, request):
        """
        Creates connection for the request and returns the created protocol.
        
        This method is a coroutine.
        
        Parameters
        ----------
        request : ``ClientRequest``
            Respective request.
        
        Returns
        -------
        protocol : ``HttpReadWriteProtocol``
        
        Raises
        ------
        ssl.SSLError
        ssl.CertificateError
        OSError
        RuntimeError
            If transport does not expose socket instance.
        """
        if request.proxy is None:
            coroutine = self.create_direct_connection(request)
        else:
            coroutine = self.create_proxy_connection(request)
        
        return await coroutine
    
    
    def get_ssl_context(self, request):
        """
        Gets ssl context for the respective request.
        
        Parameters
        ----------
        request : ``ClientRequest``
            Respective request.
        
        Returns
        -------
        ssl_context : `None | SSLContext`
            Returns `None` if the request is not `ssl`.
        """
        if not request.is_secure():
            return None
        
        ssl_context = request.ssl_context
        if (ssl_context is not None):
            return ssl_context
        
        ssl_context = self.ssl_context
        if (ssl_context is not None):
            return ssl_context
        
        return SSL_CONTEXT_UNVERIFIED
    
    
    def get_ssl_fingerprint(self, request):
        """
        Gets fingerprint for the respective request.
        
        Parameters
        ----------
        request : ``ClientRequest``
            Respective request.
        
        Returns
        -------
        fingerprint : `None | SSLFingerprint`
        """
        if not request.is_secure():
            return None
        
        ssl_fingerprint = request.ssl_fingerprint
        if (ssl_fingerprint is not None):
            return ssl_fingerprint
        
        ssl_fingerprint = self.ssl_fingerprint
        if (ssl_fingerprint is not None):
            return ssl_fingerprint
    
    
    async def create_direct_connection(self, request):
        """
        Creates a direct connection for the given request and returns the created protocol.
        
        This method is a coroutine.
        
        Parameters
        ----------
        request : ``ClientRequest``
            Respective request.
        
        Returns
        -------
        protocol : ``HttpReadWriteProtocol``
        
        Raises
        ------
        ssl.SSLError
        ssl.CertificateError
        OSError
        """
        ssl_context = self.get_ssl_context(request)
        ssl_fingerprint = self.get_ssl_fingerprint(request)
        
        causes = None
        
        async for host_info in self.resolve_host_iterator(request):
            try:
                protocol = await self.loop.create_connection_to(
                    partial_func(HttpReadWriteProtocol, self.loop),
                    host_info.host,
                    host_info.port,
                    ssl = ssl_context,
                    socket_family = host_info.family,
                    socket_protocol = host_info.protocol,
                    socket_flags = host_info.flags,
                    local_address = self.local_address,
                    server_host_name = (None if (ssl_context is None) else host_info.host_name.rstrip('.')),
                )
            except (SSLCertificateError, SSLError) as err:
                err.key = request.connection_key
                raise
            
            except OSError as exception:
                if (causes is None):
                    causes = []
                
                causes.append(exception)    
                continue
            
            if (ssl_fingerprint is not None):
                try:
                    ssl_fingerprint.check(protocol)
                except ValueError as exception:
                    protocol.close_transport(force = True)
                    
                    if (causes is None):
                        causes = []
                    
                    causes.append(exception)    
                    continue
            
            return protocol
        
        raise OSError(request.connection_key) from (None if causes is None else CauseGroup(*causes))
    
    
    async def create_proxy_connection(self, request):
        """
        Creates a proxy connection for the given request and returns the created protocol.
        
        This method is a coroutine.
        
        Parameters
        ----------
        request : ``ClientRequest``
            Respective request.
        
        Returns
        -------
        protocol : ``HttpReadWriteProtocol``
        
        Raises
        ------
        ssl.SSLError
        ssl.CertificateError
        OSError
        ProxyError
        RuntimeError
            If transport does not expose socket instance.
        """
        proxy = request.proxy
        
        proxy_request = ClientRequest(
            self.loop,
            METHOD_CONNECT,
            proxy.url,
            IgnoreCaseMultiValueDictionary(proxy.headers),
            None,
            None,
            None,
            proxy.authorization,
            request.url,
            None,
            proxy.ssl_context,
            proxy.ssl_fingerprint,
        )
        
        # create connection to proxy server
        protocol = await self.create_direct_connection(proxy_request)
        ssl_context = self.get_ssl_context(request)
        connection = Connection(self, request.connection_key.copy_proxyless(), protocol, 0)
        set_tcp_nodelay(connection.get_transport(), True)
        
        response = proxy_request.begin(connection)
        try:
            await response.start_processing()
        except:
            connection.close()
            raise
        
        try:
            connection.detach()
            response.close()
            
            try:
                status = response.status
                if status != 200:
                    reason = response.reason
                    if reason is None:
                        reason = HTTPStatus(status).phrase
                    
                    raise ProxyError(status, reason, response.headers)
            except:
                protocol.close()
                raise
            
            if not request.is_secure():
                return protocol
            
            underlying_transport = protocol._transport
            protocol = HttpReadWriteProtocol(self.loop)
            
            try:
                underlying_transport.pause_reading()
                waiter = Future(self.loop)
                ssl_protocol = SSLBidirectionalTransportLayer(
                    self.loop,
                    protocol,
                    ssl_context,
                    waiter,
                    False,
                    request.host,
                    False,
                )
                underlying_transport.set_protocol(ssl_protocol)
                ssl_protocol.connection_made(underlying_transport)
                underlying_transport.resume_reading()
                
                try:
                    await waiter
                except:
                    underlying_transport.close()
                    raise
            
            except (SSLCertificateError, SSLError) as exception:
                exception.key = request.connection_key
                raise
            
            except OSError as exception:
                raise OSError(request.connection_key, exception) from exception
            
            else:
                protocol.connection_made(ssl_protocol)
        
        except:
            response.close()
            raise
        
        return protocol



__all__ = ('LineInfo',)

from os import stat as get_stats

from .cache_constants import CONSOLE_INPUT_CACHE, FILE_INFO_CACHE, LINE_INFO_CACHE, LINE_CACHE_MAX_SIZE
from .file_info import get_file_info


class LineInfo:
    """
    Represents a cached line.
    
    Attributes
    ----------
    file_modification_time : `float`
        When the file was last modified. Using system time. Used to identify file modifications.
    file_name : `str`
        The line's owner file's name.
    file_size : `int`
        The file's size. Used to identify file modifications.
    hit : `bool`
        Whether the line is a hit.
    index : `int`
        The line's index.
    line : `str`
        The stored line. Can be an empty string.
    """
    __slots__ = ('file_modification_time', 'file_name', 'file_size', 'hit', 'index', 'line')
    
    def __new__(cls, file_name, file_size, file_modification_time, line, index, hit):
        """
        Creates a new line info.
        
        Parameters
        ----------
        file_name : `str`
            The line's owner file's name.
        file_size : `int`
            The file's size.
        file_modification_time : `float`
            When the file was last modified.
        line : `str`
            The stored line.
        index : `int`
            The line's index.
        hit : `bool`
            Whether the line is a hit.
        """
        self = object.__new__(cls)
        self.file_modification_time = file_modification_time
        self.file_name = file_name
        self.file_size = file_size
        self.hit = hit
        self.index = index
        self.line = line
        return self
    
    
    def __repr__(self):
        """Returns the line info's representation."""
        repr_parts = ['<', type(self).__name__]
        
        repr_parts.append(' file_modification_time = ')
        repr_parts.append(repr(self.file_modification_time))
        
        repr_parts.append(', file_name = ')
        repr_parts.append(repr(self.file_name))
        
        repr_parts.append(', file_size = ')
        repr_parts.append(repr(self.file_size))
        
        repr_parts.append(', hit = ')
        repr_parts.append(repr(self.hit))
        
        repr_parts.append(', index = ')
        repr_parts.append(repr(self.index))
        
        repr_parts.append(', line = ')
        repr_parts.append(repr(self.line))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def is_up_to_date(self):
        """
        Returns whether the line info is up to date.
        
        Returns
        -------
        is_up_to_date : `bool`
        """
        file_name = self.file_name
        try:
            file_info = FILE_INFO_CACHE[file_name]
        except KeyError:
            pass
        else:
            if file_info.console:
                try:
                    CONSOLE_INPUT_CACHE.move_to_end(file_name)
                except KeyError:
                    CONSOLE_INPUT_CACHE[file_name] = file_info
                return True
            
            if not file_info.alive:
                return True
            
            if self.file_size != file_info.file_size:
                return False
            
            if self.file_modification_time != file_info.file_modification_time:
                return False
            
            return True
        
        try:
            stats = get_stats(file_name)
        except OSError:
            pass
        else:
            if self.file_size != stats.st_size:
                return False
            
            if self.file_modification_time != stats.st_mtime:
                return False
            
            return True
        
        return False
    
    
    def update(self):
        """
        Updates the line info.
        """
        file_info = get_file_info(self.file_name)
        if file_info.alive:
            self.hit, self.line = file_info.get_line(self.index)
        
    
    def get(self):
        """
        Gets the line of the line info. Updates it if required.
        
        Returns
        -------
        hit : `bool`
            whether the line is a hit.
        line : `str`
            The line of the line info.
        """
        if not self.is_up_to_date():
            self.update()
        
        return self.hit, self.line


def get_file_line(file_name, line_index):
    """
    Gets the file's line.
    
    Parameters
    ----------
    file_name : `str`
        The file's name.
    line_index : `int`
        The line's index to get.
    
    Returns
    -------
    hit : `bool`
    line : `str`
    """
    key = (file_name, line_index)
    
    # Try get from cache
    try:
        line_info = LINE_INFO_CACHE[key]
    except KeyError:
        pass
    else:
        LINE_INFO_CACHE.move_to_end(key)
        return line_info.get()
    
    # Get file -> get line -> cache it
    file_info = get_file_info(file_name)
    hit, line = file_info.get_line(line_index)
    line_info = LineInfo(file_name, file_info.file_size, file_info.file_modification_time, line, line_index, hit)
    LINE_INFO_CACHE[key] = line_info
    
    if len(LINE_INFO_CACHE) >= LINE_CACHE_MAX_SIZE:
        del LINE_INFO_CACHE[next(iter(LINE_INFO_CACHE))]
    
    return hit, line



__all__ = ('FrameProxyBase',)


from ..expression_parsing import ExpressionKey


class FrameProxyBase:
    """
    Base class for frame proxies.
    
    Frame proxies are used to provide the same api for different types of frame representations.
    
    Attributes
    ----------
    alike_count : `int`
        How much alike frames are in the same exception.
    expression_info : `None | ExpressionInfo`
        Additional expression information.
    """
    __slots__ = ('alike_count', 'expression_info')
    
    def __new__(cls):
        """
        Creates a new frame proxy.
        """
        self = object.__new__(cls)
        self.alike_count = 0
        self.expression_info = None
        return self
    
    
    def __repr__(self):
        """Returns the frame proxy's representation."""
        return ''.join(['<', type(self).__name__, '>'])
    
    
    @property
    def builtins(self):
        """
        Returns the frame's builtins. Can return `None` if the variables were already omitted.
        
        Returns
        -------
        builtins : `None | dict<str, object>`
        """
        return None
    
    
    @property
    def globals(self):
        """
        Returns the frame's globals. Can return `None` if the variables were already omitted.
        
        Returns
        -------
        globals : `None | dict<str, object>`
        """
        return None
    
    
    @property
    def locals(self):
        """
        Returns the local variables of the frame. Can return `None` if the variables were already omitted.
        
        Returns
        -------
        locals : `None | dict<str, object>`
        """
        return None
    
    
    @property
    def code(self):
        """
        Returns the frame's code.
        
        Returns
        -------
        code : `None | CodeType`
        """
        return None
    
    
    @property
    def instruction_index(self):
        """
        Returns the frame's last attempted instruction index in the bytecode.
        
        Returns
        -------
        last_instruction_index : `int`
        """
        return 0
    
    
    @property
    def line_index(self):
        """
        Returns the frame's current line index in Python source code.
        
        Returns
        -------
        line_index : `int`
        """
        return 0
    
    
    @property
    def tracing_function(self):
        """
        Tracing function for the traceback's frame. Defaults to `None`.
        
        Returns
        -------
        tracing_function : `None | FunctionType`
        """
        return None
    
    
    @property
    def file_name(self):
        """
        Returns the frame's file name. Can be empty string.
        
        Returns
        -------
        file_name : `str`
        """
        code = self.code
        if code is None:
            file_name = ''
        else:
            file_name = code.co_filename
            if (file_name is None):
                file_name = ''
        
        return file_name
    
    
    @property
    def name(self):
        """
        Returns the frame's function's name. Can be empty string.
        
        Returns
        -------
        name : `str`
        """
        code = self.code
        if code is None:
            name = ''
        else:
            name = code.co_name
            if name is None:
                name = ''
        
        return name
    
    
    def __eq__(self, other):
        """Returns whether the two frame proxies are equal."""
        if not isinstance(other, FrameProxyBase):
            return NotImplemented
        
        return self._is_equal(other)
        
    
    def _is_equal(self, other):
        """
        Returns whether the two frame proxies are equal.
        
        Parameters
        ----------
        other : ``FrameProxyBase``
            The other instance.
        
        Returns
        -------
        is_equal : `bool`
        """
        if not self._is_alike(other):
            return False
        
        if self.builtins != other.builtins:
            return False
        
        if self.globals != other.globals:
            return False
        
        if self.locals != other.locals:
            return False
        
        if self.tracing_function != other.tracing_function:
            return False
        
        return True
    
    
    def __mod__(self, other):
        """Returns whether the two frame proxies are alike."""
        if not isinstance(other, FrameProxyBase):
            return NotImplemented
        
        return self._is_alike(other)
    
    
    def __rmod__(self, other):
        """Returns whether the two frame proxies are alike."""
        if not isinstance(other, FrameProxyBase):
            return NotImplemented
        
        return self._is_alike(other)
    
    
    def _is_alike(self, other):
        """
        Returns whether the two frame proxies are alike.
        
        Parameters
        ----------
        other : ``FrameProxyBase``
            The other instance.
        
        Returns
        -------
        is_alike : `bool`
        """
        if self.file_name != other.file_name:
            return False
        
        if self.line_index != other.line_index:
            return False
        
        if self.name != other.name:
            return False
        
        if self.instruction_index != other.instruction_index:
            return False
        
        return True
    
    
    @property
    def expression_key(self):
        """
        Returns the expression key for the frame.
        
        Returns
        -------
        expression_key : ``ExpressionKey``
        """
        return ExpressionKey(self.file_name, self.line_index, self.name, self.instruction_index)

    
    def has_variables(self):
        """
        Returns whether the frame proxy inherited variables on creation.
        
        Returns
        -------
        has_variables : `bool`
        """
        return False
    
    
    @property
    def line(self):
        """
        Returns the frame's line. Can be empty string. Also can contain line breaks.
        
        Returns
        -------
        line : `str`
        """
        expression_info = self.expression_info
        if expression_info is None:
            line = ''
        else:
            line = expression_info.line
        
        return line
    
    
    @property
    def lines(self):
        """
        Returns the frame's lines. Can be empty list.
        
        Returns
        -------
        lines : `list<str>`
        """
        expression_info = self.expression_info
        if expression_info is None:
            lines = []
        else:
            lines = expression_info.lines
        
        return lines



__all__ = ()

from ...utils import RichAttributeErrorBaseType

from .constants import (
    HOST_FLAG_EXTERNAL_IP_V4, HOST_FLAG_EXTERNAL_IP_V6, HOST_FLAG_EXTERNAL_NAME, HOST_FLAG_LOCAL_IP_V4,
    HOST_FLAG_LOCAL_NAME, HOST_FLAG_UNAMBIGUOUS, HOST_MASK_ANY_EXTERNAL, HOST_MASK_ANY_LOCAL, HOST_MASK_IP_ANY,
    HOST_MASK_IP_V4, HOST_MASK_IP_V6, HOST_MASK_NAME_ANY
)
from .url_host_part import URLHostPart


class URLHostInfo(RichAttributeErrorBaseType):
    """
    Represents the host information of an url.
    
    Attributes
    ----------
    _flags : `int`
        Details about how the host.
    
    _main_domain : `None | URLHostPart`
        The main domain. If the host is is an ip address or local host only this field is set. 
    
    _sub_domain : `None | URLHostPart`
        Sub-domain if any.
    
    _top_level_domain : `None | URLHostPart`
        Top level domain (TLD) if any.
    """
    __slots__ = ('_flags', '_main_domain', '_sub_domain', '_top_level_domain')
    
    
    def __new__(cls):
        """
        Use either ``.create_from_decoded`` or ``.create_from_encoded`` instead.
        
        Raises
        ------
        RuntimeError
        """
        raise RuntimeError(
            'Use either `.create_from_local_ip_v4_decoded`, `.create_from_local_ip_v4_encoded`, '
            '`.create_from_local_name_decoded`, `.create_from_local_name_encoded`, '
            '`.create_from_external_ip_v4_decoded`, `.create_from_external_ip_v4_encoded`, '
            '`.create_from_external_ip_v6_decoded`, `.create_from_external_ip_v6_encoded`, '
            '`.create_from_external_name_decoded` or `.create_from_external_name_encoded` instead.'
        )
    
    
    def __repr__(self):
        """Returns the url path's representation."""
        repr_parts = ['<', type(self).__name__]
        
        # flags
        repr_parts.append(' flags = ')
        repr_parts.append(repr(self._flags))
        
        # main_domain_decoded
        main_domain_decoded = self.main_domain_decoded
        if (main_domain_decoded is not None):
            repr_parts.append(', main_domain_decoded = ')
            repr_parts.append(repr(main_domain_decoded))
        
        # sub_domain_decoded
        sub_domain_decoded = self.sub_domain_decoded
        if (sub_domain_decoded is not None):
            repr_parts.append(', sub_domain_decoded = ')
            repr_parts.append(repr(sub_domain_decoded))
        
        # top_level_domain_decoded
        top_level_domain_decoded = self.top_level_domain_decoded
        if (top_level_domain_decoded is not None):
            repr_parts.append(', top_level_domain_decoded = ')
            repr_parts.append(repr(top_level_domain_decoded))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two url parts are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # flags
        if self._flags != other._flags:
            return True
        
        # main_domain
        if self._main_domain != other._main_domain:
            return False
        
        # sub_domain
        if self._sub_domain != other._sub_domain:
            return False
        
        # top_level_domain
        if self._top_level_domain != other._top_level_domain:
            return False
        
        return True
    
    
    @classmethod
    def create_from_local_ip_v4_decoded(cls, local_ip_v4):
        """
        Creates a new url host info from local ip (v4).
        
        Parameters
        ----------
        local_ip_v4 : `None | str`
            Local ip (v4).
        
        Returns
        -------
        self : `instance<cls>`
        """
        return cls._create_from_decoded(HOST_FLAG_LOCAL_IP_V4, local_ip_v4, None, None)
    
    
    @classmethod
    def create_from_local_ip_v4_encoded(cls, local_ip_v4):
        """
        Creates a new url host info from local ip (v4).
        
        Parameters
        ----------
        local_ip_v4 : `None | str`
            Local ip (v4).
        
        Returns
        -------
        self : `instance<cls>`
        """
        return cls._create_from_encoded(HOST_FLAG_LOCAL_IP_V4, local_ip_v4, None, None)
    
    
    @classmethod
    def create_from_local_name_decoded(cls, local_host, unambiguous):
        """
        Creates a new url host info from local (host) name.
        
        Parameters
        ----------
        local_host : `None | str`
            Local host.
        
        unambiguous : `bool`
            Whether the host is unambiguous.
        
        Returns
        -------
        self : `instance<cls>`
        """
        flags = HOST_FLAG_LOCAL_NAME
        
        if unambiguous:
            flags |= HOST_FLAG_UNAMBIGUOUS
        
        return cls._create_from_decoded(flags, local_host, None, None)
    
    
    @classmethod
    def create_from_local_name_encoded(cls, local_host, unambiguous):
        """
        Creates a new url host info from local (host) name.
        
        Parameters
        ----------
        local_host : `None | str`
            Local host.
        
        unambiguous : `bool`
            Whether the host is unambiguous.
        
        Returns
        -------
        self : `instance<cls>`
        """
        flags = HOST_FLAG_LOCAL_NAME
        
        if unambiguous:
            flags |= HOST_FLAG_UNAMBIGUOUS
        
        return cls._create_from_encoded(flags, local_host, None, None)
    
    
    @classmethod
    def create_from_external_ip_v4_decoded(cls, external_ip_v4):
        """
        Creates a new url host info from external ip (v4).
        
        Parameters
        ----------
        external_ip_v4 : `None | str`
            External ip (v4).
        
        Returns
        -------
        self : `instance<cls>`
        """
        return cls._create_from_decoded(HOST_FLAG_EXTERNAL_IP_V4, external_ip_v4, None, None)
    
    
    @classmethod
    def create_from_external_ip_v4_encoded(cls, external_ip_v4):
        """
        Creates a new url host info from external ip (v4).
        
        Parameters
        ----------
        external_ip_v4 : `None | str`
            External ip (v4).
        
        Returns
        -------
        self : `instance<cls>`
        """
        return cls._create_from_encoded(HOST_FLAG_EXTERNAL_IP_V4, external_ip_v4, None, None)
    
    
    @classmethod
    def create_from_external_ip_v6_decoded(cls, external_ip_v6):
        """
        Creates a new url host info from external ip (v6).
        
        Parameters
        ----------
        external_ip_v6 : `None | str`
            External ip (v6).
        
        Returns
        -------
        self : `instance<cls>`
        """
        return cls._create_from_decoded(HOST_FLAG_EXTERNAL_IP_V6, external_ip_v6, None, None)
    
    
    @classmethod
    def create_from_external_ip_v6_encoded(cls, external_ip_v6):
        """
        Creates a new url host info from external ip (v6).
        
        Parameters
        ----------
        external_ip_v6 : `None | str`
            External ip (v6).
        
        Returns
        -------
        self : `instance<cls>`
        """
        return cls._create_from_encoded(HOST_FLAG_EXTERNAL_IP_V6, external_ip_v6, None, None)
    
    
    @classmethod
    def create_from_external_name_decoded(cls, sub_domain, main_domain, top_level_domain, unambiguous):
        """
        Creates a new url host info from external (host) name.
        
        Parameters
        ----------
        sub_domain : `None | str`
            Sub-domain if any.
        
        main_domain : `None | str`
            The main domain.
        
        top_level_domain : `None | str`
            Top level domain (TLD) if any.
        
        unambiguous : `bool`
            Whether the host is unambiguous.
        
        Returns
        -------
        self : `instance<cls>`
        """
        flags = HOST_FLAG_EXTERNAL_NAME
        
        if unambiguous:
            flags |= HOST_FLAG_UNAMBIGUOUS
        
        return cls._create_from_decoded(flags, main_domain, sub_domain, top_level_domain)
    
    
    @classmethod
    def create_from_external_name_encoded(cls, sub_domain, main_domain, top_level_domain, unambiguous):
        """
        Creates a new url host info from external (host) name.
        
        Parameters
        ----------
        sub_domain : `None | str`
            Sub-domain if any.
        
        main_domain : `None | str`
            The main domain.
        
        top_level_domain : `None | str`
            Top level domain (TLD) if any.
        
        unambiguous : `bool`
            Whether the host is unambiguous.
        
        Returns
        -------
        self : `instance<cls>`
        """
        flags = HOST_FLAG_EXTERNAL_NAME
        
        if unambiguous:
            flags |= HOST_FLAG_UNAMBIGUOUS
        
        return cls._create_from_encoded(flags, main_domain, sub_domain, top_level_domain)
    
    
    @classmethod
    def _create_from_decoded(cls, flags, main_domain, sub_domain, top_level_domain):
        """
        Creates a new url host info from decoded values.
        
        Parameters
        ----------
        flags : `int`
            Details about how the host.
        
        main_domain : `None | str`
            The main domain. If the host is is an ip address or local host only this field is set. 
        
        sub_domain : `None | str`
            Sub-domain if any.
        
        top_level_domain : `None | str`
            Top level domain (TLD) if any.
        
        Returns
        -------
        self : `instance<cls>`
        """
        if main_domain is None:
            main_domain = None
        else:
            main_domain = URLHostPart.create_from_decoded(main_domain)
        
        if sub_domain is None:
            sub_domain = None
        else:
            sub_domain = URLHostPart.create_from_decoded(sub_domain)
        
        if top_level_domain is None:
            top_level_domain = None
        else:
            top_level_domain = URLHostPart.create_from_decoded(top_level_domain)
        
        self = object.__new__(cls)
        self._flags = flags
        self._main_domain = main_domain
        self._sub_domain = sub_domain
        self._top_level_domain = top_level_domain
        return self
    
    
    @classmethod
    def _create_from_encoded(cls, flags, main_domain, sub_domain, top_level_domain):
        """
        Creates a new url host info from encoded values.
        
        Parameters
        ----------
        flags : `int`
            Details about how the host.
        
        main_domain : `None | str`
            The main domain. If the host is is an ip address or local host only this field is set. 
        
        sub_domain : `None | str`
            Sub-domain if any.
        
        top_level_domain : `None | str`
            Top level domain (TLD) if any.
        
        Returns
        -------
        self : `instance<cls>`
        """
        if main_domain is None:
            main_domain = None
        else:
            main_domain = URLHostPart.create_from_encoded(main_domain)
        
        if sub_domain is None:
            sub_domain = None
        else:
            sub_domain = URLHostPart.create_from_encoded(sub_domain)
        
        if top_level_domain is None:
            top_level_domain = None
        else:
            top_level_domain = URLHostPart.create_from_encoded(top_level_domain)
        
        self = object.__new__(cls)
        self._flags = flags
        self._main_domain = main_domain
        self._sub_domain = sub_domain
        self._top_level_domain = top_level_domain
        return self
    
    
    @property
    def main_domain_decoded(self):
        """
        Returns the host info's decoded main-domain.
        
        Returns
        -------
        main_domain : `None | str`
        """
        main_domain = self._main_domain
        if (main_domain is not None):
            return main_domain.decoded
    
    
    @property
    def sub_domain_decoded(self):
        """
        Returns the host info's decoded sub-domain.
        
        Returns
        -------
        main_domain : `None | str`
        """
        sub_domain = self._sub_domain
        if (sub_domain is not None):
            return sub_domain.decoded

    
    @property
    def top_level_domain_decoded(self):
        """
        Returns the host info's decoded top-level-domain.
        
        Returns
        -------
        main_domain : `None | str`
        """
        top_level_domain = self._top_level_domain
        if (top_level_domain is not None):
            return top_level_domain.decoded
    
    
    @property
    def main_domain_encoded(self):
        """
        Returns the host info's encoded main-domain.
        
        Returns
        -------
        main_domain : `None | str`
        """
        main_domain = self._main_domain
        if (main_domain is not None):
            return main_domain.encoded
    
    
    @property
    def sub_domain_encoded(self):
        """
        Returns the host info's encoded sub-domain.
        
        Returns
        -------
        main_domain : `None | str`
        """
        sub_domain = self._sub_domain
        if (sub_domain is not None):
            return sub_domain.encoded
    
    
    @property
    def top_level_domain_encoded(self):
        """
        Returns the host info's encoded top-level-domain.
        
        Returns
        -------
        main_domain : `None | str`
        """
        top_level_domain = self._top_level_domain
        if (top_level_domain is not None):
            return top_level_domain.encoded
    
    
    def is_ip_v4(self):
        """
        Returns whether the host is an ip v4 address.
        
        Returns
        -------
        is_ip_v4 : `bool`
        """
        return True if self._flags & HOST_MASK_IP_V4 else False
    
    
    def is_ip_v6(self):
        """
        Returns whether the host is an ip v6 address.
        
        Returns
        -------
        is_ip_v6 : `bool`
        """
        return True if self._flags & HOST_MASK_IP_V6 else False
    
    
    def is_ip(self):
        """
        Returns whether the host is an ip address.
        
        Returns
        -------
        is_ip : `bool`
        """
        return True if self._flags & HOST_MASK_IP_ANY else False

    
    def is_name(self):
        """
        Returns whether the host is a (host) name.
        
        Returns
        -------
        is_name : `bool`
        """
        return True if self._flags & HOST_MASK_NAME_ANY else False
    
    
    def is_local(self):
        """
        Returns whether the host referenced a local address or (host) name.
        
        Returns
        -------
        is_local : `bool`
        """
        return True if self._flags & HOST_MASK_ANY_LOCAL else False
    
    
    def is_external(self):
        """
        Returns whether the host referenced an external address or (host) name.
        
        Returns
        -------
        is_external : `bool`
        """
        return True if self._flags & HOST_MASK_ANY_EXTERNAL else False
    
    
    def is_unambiguous(self):
        """
        Returns whether the host is unambiguous.
        
        Returns
        -------
        is_unambiguous : `bool`
        """
        return True if self._flags & HOST_FLAG_UNAMBIGUOUS else False
